/**
 * 
 * RoadHog Compiler
 * v0.9
 * (c) 2001 Frederik Ramm 
 * Released under GPL
 *
 * I know this needs documentation. 
 * See http://www.remote.org/frederik/roadhog/software/ for details.
 */

package org.remote.roadhog;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;
import magick.*;

import gnu.getopt.Getopt;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import org.apache.xerces.parsers.DOMParser;

public class Compiler
{

	////////////////////////////////
	// @@ Properties
	////////////////////////////////

	private PrintStream reportingStream;		// where we log progress
	private String destinationPath = ".";		// where we write our results
	private boolean quickAndDirty = false;		// no contrast/brightness alteration 
	private boolean mapsOnly = false;
	private boolean incremental = true;
	private boolean vertexOnly = false;

	////////////////////////////////
	// @@ private data
	////////////////////////////////

	private HashMap journeys = new HashMap ();
	private HashMap vertices = new HashMap ();
	private HashMap edges = new HashMap ();
	private HashMap maps = new HashMap ();
	private Vector otherElements = new Vector ();
	private DOMParser parser = new DOMParser ();
	private Vector sizeSpecs = new Vector ();
	private int magickUsage = 0;

	public final String version = "v0.9";

	////////////////////////////////
	// @@ Constructor
	////////////////////////////////

	public Compiler ()
	{
/* old
		sizeSpecs.add (new SizeSpec (256, 192, 40, 25, 256, "tiny"));
		sizeSpecs.add (new SizeSpec (400, 300, 50, 40, 256, "small"));
		sizeSpecs.add (new SizeSpec (512, 384, 60, 60, 256, "medium"));
		sizeSpecs.add (new SizeSpec (800, 600, 80, 80, 256, "large"));
//		sizeSpecs.add (new SizeSpec (1024, 768, 80, 100, 256, "huge"));
*/
		sizeSpecs.add (new SizeSpec (256, 192, 35, 20, 32, "tiny"));
		sizeSpecs.add (new SizeSpec (400, 300, 45, 35, 64, "small"));
		sizeSpecs.add (new SizeSpec (512, 384, 50, 45, 128, "medium"));
		sizeSpecs.add (new SizeSpec (720, 540, 70, 60, 256, "large"));
		sizeSpecs.add (new SizeSpec (1024, 768, 80, 0, 0, "huge"));

	}

	////////////////////////////////
	// @@ Public Methods
	////////////////////////////////

	public void readFile (String filename)
	throws ParseException, IOException, SAXException
	{
		XMLUtils.readFile (parser, filename, vertices, maps, edges, journeys, otherElements, reportingStream, false, true);
	}

	public void produceOutput ()
	throws IOException
	{
		message ("data objects loaded: " + vertices.size () + "v " + edges.size () + "e " +
			journeys.size () + "j " + maps.size () + "m " + otherElements.size () + "o");

		report ("Producing output...");

		// generate lookup file for coordinates
		if (!vertexOnly) createLocationIndex (destinationPath + "/control/location.txt");

		// generate lookup file for vias
		if (!vertexOnly) createPlaceIndex (destinationPath + "/control/places.txt");

		// pre-process the maps
		// TODO: only process those references from an edge!
		for (Iterator i = maps.values ().iterator (); i.hasNext (); )
		{
			Map m = (Map) i.next ();
			processMap (m, destinationPath + "/maps/");
		}

		// edges will automatically process associated journeys
		if (!vertexOnly) for (Iterator i = edges.values ().iterator (); i.hasNext (); )
		{
			Edge e = (Edge) i.next ();
			processEdge (e, destinationPath + "/img/", destinationPath + "/control/");
		}
		
		// post-process the maps
		if (!vertexOnly) for (Iterator i = maps.values ().iterator (); i.hasNext (); )
		{
			Map m = (Map) i.next ();
			createDotMaps (m, destinationPath + "/maps/");
		}

		if (!mapsOnly) for (Iterator i = vertices.values ().iterator (); i.hasNext (); )
		{
			Vertex v = (Vertex) i.next ();
			if (v.hasDescription ())
			{
				createVertexPage (v, destinationPath + "/places/", "../img/", "../maps/", "../roadhog.cgi");
			}
		}

		if (!mapsOnly)
		{
			createIndexPage (destinationPath + "/", "places/", (Map) maps.get ("mini"));
			createShowcasePage (destinationPath + "/places/");
		}

		report ("Output finished.\n");

	}

	private void createShowcasePage (String destDir)
	{

		int counter = 0;
		
		try
		{
			PrintWriter p = new PrintWriter (new FileOutputStream (destDir + "showcase.html"));

			p.println ("<html><head>");
			p.print ("<title>On the Road in North-West Scotland: Showcase");
			p.println ("</title>");
			p.println (Html.getStyleTag ());
			p.println ("</head><body>");

			p.println (Html.makeHeadTable ("Showcase", true));

			p.println ("<p><table width=100%><tr><td colspan=3>This is a selection of ");
			p.println ("cool, spooky, funny, beautiful or just peculiar sights you'll encounter ");
			p.println ("when you \"drive\" through the web experience. Links are provided ");
			p.println ("to take you directly to the journey from which each photo was taken.");
			p.println ("<p></td></tr>");
			p.println ("<tr><td>&nbsp;</td></tr>");

			for (Enumeration e = otherElements.elements (); e.hasMoreElements (); )
			{
				Object o = e.nextElement ();
				if (o instanceof ShowcaseItem)
				{
					ShowcaseItem item = (ShowcaseItem) o;

					p.print ("<tr valign=top><td");
					if (item.illustrations.size () == 2)
					{
						p.print (" colspan=3");
					}
					p.println (">");
					for (Enumeration n = item.illustrations.elements (); n.hasMoreElements (); )
					{
						Illustration ill = (Illustration) n.nextElement ();
						counter++;
						String imgname = "showcase" + counter + ".jpg";
						ill.saveAs (destDir + imgname);
						p.print ("<img border=2 src=\"");
						p.print (imgname);
						p.print ("\" width=");
						p.print (ill.size.width);
						p.print (" height=");
						p.print (ill.size.height);
						p.print (">");
						if (item.illustrations.size () == 2)
						{
							p.println ("&nbsp;&nbsp;&nbsp;");
						}
						else
						{
							p.println ("<br>");
						}
					}
					if (item.illustrations.size () == 2)
					{
						p.println ("<p>");
					}
					else
					{
						p.println ("</td><td>&nbsp;</td><td>");
					}
					p.println (item.getText ("en"));
					// p.println ("<p>");
					
					for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
					{
						Journey j = (Journey) i.next ();
						if (item.journeyMatches (j))
						{
							p.print ("<a href=\"../roadhog.cgi?show=");
							p.print (j.getHandle ());
							p.print ("\">");
							p.print (j.from.name);
							p.print (" to ");
							p.print (j.to.name);
							p.print ("</a>");
							break;
						}
					}
					p.println ("</td></tr>\n<tr><td>&nbsp;</td></tr>");
				}
			}
			p.println ("</table><p>");

			String left = "<a href=\"../index.html\">Place List</a> &nbsp; <a href=\"../search.cgi\">Search</a> &nbsp; <a href=\"../faq.html\">Help &amp; FAQ</a>";
			String right = version;
			p.println (Html.makeNavTable (left, right));
			p.println ("</body></html>");
			p.close ();
		}
		catch (Exception x)
		{
			x.printStackTrace ();
		}
	}


	private void createIndexPage (String destDir, String placeDir, Map minimap)
	{
		if (minimap == null)
		{
			report ("no mini map found - not creating map overview");
			return;
		}

		new File (destDir).mkdirs ();
		StringBuffer clickableMap = new StringBuffer ();
		Vector redDotPlaces = new Vector ();
		int photoCounter = 0;
		
		try
		{
			ImageInfo mapImageInfo = new ImageInfo (minimap.filename);
			report ("  loading " + minimap.filename);
			MagickImage mapImage1 = new MagickImage (mapImageInfo);
			MagickImage mapImage2 = mapImage1.cloneImage (0, 0, true);
			Point pixel = new Point ();

			for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
			{
				Journey j = (Journey) i.next ();
				if (j.photo != null) 
				{
					photoCounter += j.photo.length;
					for (int t=0; t<j.photo.length; t++)
					{
						Waypoint w = j.photo[t].waypoint;
						if (w == null) continue;
						minimap.mapGeoToPixel (w.coord, pixel);
						minimap.applyCorrection (pixel);
						if (w.hack != null)
						{
							int perc = minimap.lastUsedCorrectionPercent;
							pixel.translate (w.hack.x * perc / 100, w.hack.y * perc / 100);
						}
						int radius = 3;
						DrawInfo di = new DrawInfo (mapImageInfo);
						di.setPrimitive ("fill black circle "+pixel.x+","+pixel.y+" "+
							(pixel.x+radius)+","+(pixel.y+radius));
						mapImage1.drawImage (di);
						/*
						if ((w.mapPixel.x > 300) && (w.mapPixel.y < 250))
							System.out.println ("strange pos for j from "+j.from.id + " to "+j.to.id+" image #"+t);
							*/
					}
				}
			}

			for (Iterator i = vertices.values ().iterator (); i.hasNext (); )
			{
				Vertex v = (Vertex) i.next ();
				if (v.hasDescription ())
				{
					Vector going = new Vector ();
					for (int c=0; c<v.connector.length; c++)
					{
						Edge e = findEdge (v, c);
						if (e != null)
						{
							findJourneys (e, v, going, null);
							if (going.size () > 0) break;
						}
					}

					Waypoint w = null;
					for (int g=0; g<going.size (); g++)
					{
						Journey j = (Journey) going.elementAt (g);
						if (j.photo != null)
						{
							w = j.photo[0].waypoint;
							if (w != null) break;
						}
					}
					if (w == null) continue;

					minimap.mapGeoToPixel (w.coord, pixel);
					minimap.applyCorrection (pixel);

					int radius = 3;

					clickableMap.append ("<area href=\"");
					clickableMap.append (placeDir);
					clickableMap.append (v.getBaseName ());
					clickableMap.append (".html\" shape=circle coords=\"");
					clickableMap.append (pixel.x);
					clickableMap.append (",");
					clickableMap.append (pixel.y);
					clickableMap.append (",8 alt=\"");
					clickableMap.append (v.name);
					clickableMap.append (">\n");
					redDotPlaces.addElement (v);

					DrawInfo di = new DrawInfo (mapImageInfo);
					di.setPrimitive ("fill black circle "+pixel.x+","+pixel.y+" "+
						(pixel.x+radius)+","+(pixel.y+radius));
					mapImage1.drawImage (di);
					di.setPrimitive ("fill black circle "+pixel.x+","+pixel.y+" "+
						(pixel.x+radius)+","+(pixel.y+radius));
					mapImage2.drawImage (di);
					di.setPrimitive ("fill red circle "+pixel.x+","+pixel.y+" "+
						(pixel.x+radius-1)+","+(pixel.y+radius-1));
					mapImage1.drawImage (di);
					di.setPrimitive ("fill red circle "+pixel.x+","+pixel.y+" "+
						(pixel.x+radius-1)+","+(pixel.y+radius-1));
					mapImage2.drawImage (di);

				}
			}
			mapImage1.setFileName (destDir + "img/index1.gif");
			report ("  writing " + destDir+"img/index1.gif");
			mapImage1.writeImage (mapImageInfo);
			mapImage2.setFileName (destDir + "img/index2.gif");
			report ("  writing " + destDir+"img/index2.gif");
			mapImage2.writeImage (mapImageInfo);
		}
		catch (MagickException mx)
		{
			mx.printStackTrace ();
		}

		File path = new File (destDir + "index.html");
		report ("  creating " + destDir + "index.html");
		NumberFormat nf = NumberFormat.getInstance ();
		try
		{
			PrintWriter o = new PrintWriter (new FileOutputStream (path));

			o.println ("<html><head>");
			o.print ("<title>On the Road in North-West Scotland");
			o.println ("</title>");
			o.println ("<meta name=\"keywords\" content=\"scotland,highlands,pictures,photos,maps,GPS\">";
			o.print ("<meta name=\"description\" content=\"");
			o.print (nf.format (photoCounter));
			o.println ("photos of the Scottish Highland, complete with GPS co-ordinates, taken while driving.\">");
			o.println (Html.getStyleTag ());
			o.println ("</head><body>");

			o.println (Html.makeHeadTable ("On the Road in North-West Scotland", true));

			o.println ("<p><table width=80% align=center><tr><td align=center>");
			o.println ("<font size=+1>Oh dear! This is what happens when a mad tourist ");
			o.println ("straps a camera and a GPS to his Land Rover and shoots ");
			o.print (nf.format (photoCounter));
			o.println (" photos.<sup>*)</sup></font>");
			o.println ("</td></tr><tr><td>&nbsp;</td></tr>");
			o.println ("<tr><td align=center>");
			
			o.println ("<img src=\"img/index.gif\" usemap=\"#index.map\" border=0 align=center>");

			o.println ("<p>(Click on one of the red dots for a starting point,");
			o.println ("<br>or select from the list below.)");
			o.println ("</td></tr>");

			o.println ("</table><p>");

			Collections.sort (redDotPlaces, new Comparator () {
				public int compare (Object o1, Object o2) {
					return ((Vertex)o1).name.compareTo (((Vertex)o2).name);
				}
			});

			o.print   ("<table align=center cellpadding=0 cellspacing=0 border=0 bgcolor=#");
			o.print   (Html.NAV_TABLE_BACKGROUND);
			o.println ("><tr><td colspan=9>&nbsp;</td></tr>");
			int step = (redDotPlaces.size () + 3) / 4;
			for (int i=0; i<step; i++)
			{
				o.print   ("<tr>");
				for (int j=i; j < i+4*step; j+= step)
				{
					if (j == i) o.print ("<td>&nbsp;&nbsp;&nbsp;&nbsp;</td>");
					if (j < redDotPlaces.size ())
					{
						Vertex v = (Vertex) redDotPlaces.elementAt (j);
						o.print ("\t<td><a href=\"");
						o.print (placeDir);
						o.print (v.getBaseName ());
						o.print (".html\">");
						o.print (Html.insertNonBreakingSpaces (v.name));
						o.println ("</a></td>");
					}
					else
					{
						o.print ("\t<td>&nbsp;</td>");
					}
					o.print ("<td>&nbsp;&nbsp;&nbsp;&nbsp;</td>");
				}
				o.println ("</tr>");
			}
			o.println ("<tr><td colspan=9>&nbsp;</td></tr></table><p><table width=100%>");
			o.print   ("<tr><td>Have a look at the <a href=\"places/showcase.html\">showcase page</a> ");
			o.println ("and catch a glimpse of what to expect!<p> ");

			o.print   ("Alternatively, there's a <a href=\"journeys.html\">list of all ");
			o.print   (journeys.size ());
			o.print   (" individual journey parts,</a> and an exhaustive <a href=\"search.cgi\">");
			o.println ("search facility</a> that lets you search for place names ");
			o.print   ("as well as for co-ordinates. For everything else, see the <a href=\"faq.html\">");
			o.println ("Help&nbsp;&amp;&nbsp;FAQ document.</a>");
			o.println ("</table>");

			o.println ("<table width=100%><tr><td><font size=-1>*) The actual number is almost ");
			o.print   ("twice that but only ");
			o.print   (nf.format (photoCounter));
			o.println (" have made it onto the web.</font></td></tr></table>");
			o.println ("<map name=\"index.map\">");
			o.print   (clickableMap);
			o.println ("</map>");
			o.println (Html.makeNavTable ("(c) 2001 <a href=\"http://www.remote.org/frederik/contact.html\">Frederik Ramm</a>", version + " - All rights reserved."));
			o.println ("</body></html>");
			o.close ();
		}
		catch (Exception x)
		{
			x.printStackTrace ();
		}

		path = new File (destDir + "journeys.html");
		report ("  creating " + destDir + "journeys.html");
		try
		{
			PrintWriter o = new PrintWriter (new FileOutputStream (path));

			o.println ("<html><head>");
			o.print ("<title>On the Road in North-West Scotland: Journey Segments");
			o.println ("</title>");
			o.println (Html.getStyleTag ());
			o.println ("</head><body>");

			o.println (Html.makeHeadTable ("Journey Segments", true));
			o.println ("<p><table width=100%><tr><td>");

			o.println ("This page lists the journey segments for each \"node\" (a \"node\"");
			o.println ("usually being a place where you have a choice of roads, read: ");
			o.println ("an intersection). Not every place I've been through is considered");
			o.println ("a \"node\", so if you're looking for a small place that is not");
			o.println ("in this list, try the <a href=\"search.cgi\">search facility.</a>");
			o.println ("</td></tr></table><p><table>");


			Vector vertList = new Vector (vertices.values ());
			Collections.sort (vertList, new Comparator () {
				public int compare (Object o1, Object o2) {
					return ((Vertex)o1).name.compareTo (((Vertex)o2).name);
				}
			});

			int step = (vertList.size () + 1) / 2;
			for (int i = 0; i < step; i++)
			{
				o.print   ("<tr bgcolor=");
				o.print   (Html.NAV_TABLE_BACKGROUND);
				o.println (">");
				for (int j = i; j < vertList.size (); j+=step)
				{
					Vertex v = (Vertex) vertList.elementAt (j);
					if (j > i) o.print ("<td width=2%></td>");
					o.print ("<td colspan=2 width=49%>");
					o.print (v.name);
					o.print ("</td>");
				}
				o.println ("</tr>");

				o.println ("<tr valign=top>");
				for (int j = i; j < vertList.size (); j+=step)
				{
					Vertex v = (Vertex) vertList.elementAt (j);
					if (j > i) o.print ("<td width=2%>&nbsp;</td>");
					o.print ("<td>&nbsp;</td><td>");

					for (int c=0; c<v.connector.length; c++)
					{
						Vector jlist[] = new Vector[2];
						jlist[0] = new Vector ();
						jlist[1] = new Vector ();
						Edge e = findEdge (v, c);
						if (e != null)
						{
							findJourneys (e, v, jlist[0], jlist[1]);
						}
						for (int jlc=0; jlc<2; jlc++)
						{
							int n=0;
							for (Enumeration m = jlist[jlc].elements(); m.hasMoreElements (); )
							{
								Journey jj = (Journey) m.nextElement ();
								o.print ("<a href=\"roadhog.cgi?show=");
								o.print (jj.getHandle ());
								o.print ("\">");
								o.print (jj.from.name);
								o.print (" to ");
								o.print (jj.to.name);
								o.print ("</a>");
								if (n++ != 0)
								{
									o.print (" <font size=-1>(alt. version");
									String det = jj.getDetail ("en");
									if (det != null)
									{
										o.print (" - ");
										o.print (det);
									}
									o.print (")</font>");
								}
								o.println ("<br>");
							}
						}
					}
					o.print ("</td>");
				}
				o.println ("</tr>");
			}

			o.println ("</table><p>");

			String left = "<a href=\"index.html\">Place List</a> &nbsp; <a href=\"search.cgi\">Search</a> &nbsp; <a href=\"faq.html\">Help &amp; FAQ</a>";
			String right = version;
			o.println (Html.makeNavTable (left, right));
			o.println ("</body></html>");
			o.close ();
		}
		catch (Exception x)
		{
			x.printStackTrace ();
		}
	}


	private void createVertexPage (Vertex v, String destDir, String imgDir, String mapDir, String scriptURL)
	throws IOException
	{
		String size = "tiny";
		String basename = v.getBaseName ();

		new File (destDir).mkdirs ();
		File path = new File (destDir + basename + ".html");
		NumberFormat nf = NumberFormat.getInstance ();
		nf.setParseIntegerOnly (true);
		
		report ("  creating " + destDir + basename + ".html");

		PrintWriter o = new PrintWriter (new FileOutputStream (path));
		SizeSpec spec = null;
		for (Enumeration i = sizeSpecs.elements (); i.hasMoreElements (); )
		{
			spec = (SizeSpec) i.nextElement ();
			if (spec.name.equals (size)) break;
		}

		o.println ("<html><head>");
		o.print ("<title>On the Road in North-West Scotland: ");
		o.print (v.name);
		o.println ("</title>");
		o.println (Html.getStyleTag ());
		o.println ("</head><body>");

		o.println (Html.makeHeadTable (v.name, true));
		String left = "<a href=\"../\">Place List</a> &nbsp; <a href=\"../search.cgi\">Search</a> &nbsp; <a href=\"../faq.html\">Help &amp; FAQ</a>";
		String right = version;
		o.println (Html.makeNavTable (left, right));

		o.println ("<table width=100% cellpadding=0 cellspacing=5><tr valign=top><td>");
		String desc = v.getDescription ("en");
		if (desc != null)
		{
			o.println (desc);
		}
		// create images for illustrations
		Vector illu = v.illustrations;
		if (illu != null)
		{
			int ilcnt = 0;
			for (Enumeration e = illu.elements (); e.hasMoreElements (); )
			{
				Illustration il = (Illustration) e.nextElement ();
				ilcnt++;
				String target = destDir + basename + ilcnt + ".jpg";
				if (! (incremental && fileExists (target)))
				{
					report ("  creating " + target);
					try
					{
						il.saveAs (target);
					}
					catch (MagickException mx)
					{
						mx.printStackTrace ();
					}
				}
			}

		}

		Edge e = null;
		Point mapPixel = new Point (0,0);;
		int mapPixelCount = 0;
		StringBuffer conn = new StringBuffer ();
		StringBuffer connList = new StringBuffer ();

		for (int i=0; i<v.connector.length; i++)
		{
			Vector coming = new Vector ();
			Vector going = new Vector ();
			e = findEdge (v, i);
			if (e == null) continue;
			findJourneys (e, v, going, coming);

			Journey c = null;
			Journey g = null;
			if (coming.size () > 0) c = (Journey) coming.elementAt (0);
			if (going.size () > 0) 
			{
				g = (Journey) going.elementAt (0);
				int count = 0;
				for (Enumeration n = going.elements (); n.hasMoreElements (); )
				{
					Journey j = (Journey) n.nextElement ();
					connList.append ("<li><a href=\"");
					connList.append (scriptURL);
					connList.append ("?show=");
					connList.append (j.getHandle ());
					connList.append ("\">Drive ");
					connList.append (headingToWord (v.direction[i], "en"));
					connList.append ("</a> towards ");
					connList.append (j.to.getUltimateDestination (e));
					if (count++ > 0)
					{
						connList.append (" (alternative version");
						String det = j.getDetail ("en");
						if (det != null)
						{
							connList.append (" - ");
							connList.append (det);
						}
						connList.append (")");
					}
					connList.append ("\n");
				}
			}

			if ((g == null) && (c == null)) continue;

			if (g != null)
			{
				conn.append ("<img border=2 src=\"");
				conn.append (imgDir);
				nf.setMinimumIntegerDigits ((g.photo.length < 100) ? 2 : 3);
				conn.append (g.getHandle () + "/" + spec.getLetter () + "/");
				conn.append (nf.format (1));
				conn.append (".jpg\" width=");
				conn.append (spec.imageX);
				conn.append (" height=");
				conn.append (spec.imageY);
				conn.append ("> &nbsp;\n");
				g.photo[0].waypoint.determineMapPixel (e.map);
				mapPixel.translate (g.photo[0].waypoint.mapPixel.x,
					g.photo[0].waypoint.mapPixel.y);
				mapPixelCount++;
			}
			if (c != null)
			{
				conn.append ("<img border=2 src=\"");
				conn.append (imgDir);
				nf.setMinimumIntegerDigits ((c.photo.length < 100) ? 2 : 3);
				conn.append (c.getHandle () + "/" + spec.getLetter () + "/");
				conn.append (nf.format (c.photo.length));
				conn.append (".jpg\" width=");
				conn.append (spec.imageX);
				conn.append (" height=");
				conn.append (spec.imageY);
				conn.append ("\">\n");
				c.photo[c.photo.length-1].waypoint.determineMapPixel (e.map);
				mapPixel.translate (c.photo[c.photo.length-1].waypoint.mapPixel.x,
					c.photo[c.photo.length-1].waypoint.mapPixel.y);
				mapPixelCount++;
			}

			conn.append ("<br>\n");
			if (g != null)
			{
				conn.append ("<a href=\"");
				conn.append (scriptURL);
				conn.append ("?show=");
				conn.append (g.getHandle ());
				conn.append ("\">");
			}
			conn.append ("Connecting ");
			conn.append (headingToWord (v.direction[i], "en"));

			if ((e.road != null) && !e.road.equals (""))
			{
				conn.append (" on the ");
				conn.append (e.road);
			}

			if (g != null)
			{
				conn.append (" towards ");
				conn.append (g.to.getUltimateDestination (e));
			}
			if (g != null)
			{
				conn.append ("</a>");
			}
			conn.append ("<br>\n");
			if ((g != null) && (c != null))
			{
				conn.append ("(Photo left: going there; right: coming into ");
				conn.append (v.name);
				conn.append (" from there)\n");
			}
			else if (g == null)
			{
				conn.append ("(Photo: coming from there)\n");
			}
			else
			{
				conn.append ("(Photo: going there)\n");
			}

			conn.append ("<p>\n");
		}

		if (connList.length () > 0)
		{
			o.println ("<ul>");
			o.println (connList);
			o.println ("</ul>");
			// o.println ("<em>(See below for photos.)</em>");
		}

		// check for hyperlinks
		Vector links = v.hyperlinks;
		if (links != null)
		{
			o.println ("<strong>External Links</strong>");
			o.println ("<ul>");
			for (Enumeration n = links.elements (); n.hasMoreElements (); )
			{
				Hyperlink hy = (Hyperlink) n.nextElement ();
				o.print ("<li><a href=\"");
				o.print (hy.getUrl ("en"));
				o.print ("\">");
				o.print (hy.getText ("en"));
				o.println ("</a>");
			}
			o.println ("</ul>");
		}

		if (illu != null)
		{
			Illustration il = (Illustration) illu.elementAt (0);
			o.println ("</td><td>&nbsp;</td>");
			o.println ("<td align=right>");
			o.print ("<img border=2 src=\"");
			o.print (basename);
			o.println ("1.jpg\" width=" + il.size.width + " height=" + il.size.height +">");
			o.println ("<br><font size=-1><em>");
			o.println (il.getText ("en"));
			o.println ("</em></font>");
		}
		o.println ("</td></tr></table>");
		for (Enumeration i = sizeSpecs.elements (); i.hasMoreElements (); )
		{
			spec = (SizeSpec) i.nextElement ();
			if (spec.name.equals ("medium")) break;
		}

		if (mapPixelCount > 0)
		{
			o.println (Html.makeHeadTable ("Map", false));
			o.println ("<table align=center border=1><tr><td>");

			MapInfo mi = (MapInfo) e.map.mapInfo.get ("medium");
			Point scaledPos = new Point (mapPixel.x * spec.mapScale / mapPixelCount / 100, 
				mapPixel.y * spec.mapScale / mapPixelCount / 100);

			int startRow = -100, startCol = -100;


			for (int row=0; (startRow == -100 && (row<mi.rows)); row++)
			for (int col=0; (startRow == -100 && (col<mi.columns)); col++)
			{
				if (mi.partRect[row][col].contains (scaledPos))
				{
					startRow = row - 1;
					startCol = col - 3;
				}
			}
			
			for (int row = 0; row < 3; row++)
			{
				for (int col = 0; col < 7; col++)
				{
					o.print ("<img src=\"");
					o.print (mapDir);
					o.print (e.map.id + "/l/");
					o.print (row + startRow + 1);
					o.print ("x");
					o.print (col + startCol + 1);
					o.print (".gif\">");
				}
				o.println ("<br>");
			}
			o.println ("</td></tr></table>");
		}

		o.println ("<br>");
		o.println (Html.makeHeadTable ("Connections", false));
		o.println ("<table align=center><tr><td>");
		o.print (conn);
		o.println ("</td></tr></table>");

		left = "<a href=\"../\">Place List</a> &nbsp; <a href=\"../search.cgi\">Search</a> &nbsp; <a href=\"../faq.html\">Help &amp; FAQ</a>";
		right = version;
		o.println (Html.makeNavTable (left, right));

		o.println ("</body></html>");
		o.close ();
			 
	}

	private Edge findEdge (Vertex v, int conn)
	{
		for (Iterator j = edges.values ().iterator (); j.hasNext (); )
		{
			Edge e = (Edge) j.next ();
			if (((e.vertex[0] == v) && (e.connector[0].equals (v.connector[conn]))) || 
				((e.vertex[1] == v) && (e.connector[1].equals (v.connector[conn])))) return e;
		}
		return null;
	}

	private void findJourneys (Edge e, Vertex reference, Vector going, Vector coming)
	{
		for (Iterator k = journeys.values ().iterator (); k.hasNext (); )
		{
			Journey j = (Journey) k.next ();
			if (((j.from == e.vertex[0]) && (j.to == e.vertex[1])) ||
				((j.from == e.vertex[1]) && (j.to == e.vertex[0])))
			{
				if ((j.getAlternative () == null) || (j.getAlternative ().equals (e.alternative)))
				{
					if (j.from == reference)
					{
						if (going != null) going.addElement (j);
					}
					else
					{
						if (coming != null) coming.addElement (j);
					}
				}
			}
		}
		Comparator priorityComp = new Comparator () {
			public int compare (Object a, Object b) {
				return (((Journey)a).priority > ((Journey)b).priority) ? -1 :
					(((Journey)a).priority < ((Journey)b).priority) ? 1 : 0;
			}
		};

		if (coming != null) Collections.sort (coming, priorityComp);
		if (going != null) Collections.sort (going, priorityComp);

	}

	////////////////////////////////
	// @@ Properties
	////////////////////////////////

	public void setReportingStream (PrintStream str)
	{
		reportingStream = str;
	}
	public PrintStream getReportingStream ()
	{
		return reportingStream;
	}

	public void setDestinationPath (String path)
	{
		destinationPath = path;
	}
	public String getDestinationPath ()
	{
		return destinationPath;
	}

	public void setQuickAndDirty (boolean qd)
	{
		quickAndDirty = qd;
	}
	public boolean getQuickAndDirty ()
	{
		return quickAndDirty;
	}
		

	////////////////////////////////
	// @@ Helpers
	////////////////////////////////

	private void report (String message)
	{
		if (reportingStream != null) reportingStream.println (message);
	}
	private void message (String message)
	{
		System.out.println (message);
	}

	private void processEdge (Edge edge, String staticDir, String dynamicDir)
	throws IOException
	{
		String handle = edge.getHandle ();
		report ("  processing Edge: " + handle + "...");
		
		String staticEdgeDir = staticDir;
		String dynamicEdgeDir = dynamicDir;
		// HashMap journeyList = new HashMap ();

		// this processes all the photos
		for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
		{
			Journey j = (Journey) i.next ();
			if (j.edgeMatches (edge))
			{
				// journeyList.put (add, j.description);
				processJourney (j, edge, staticEdgeDir + j.getHandle () + "/", dynamicEdgeDir + j.getHandle () + "/");
			}
		}
		
		
	}

	public void createPlaceIndex (String filename)
	throws IOException
	{
		// this will be a hash map whose elements are hash sets of edges
		HashMap vias = new HashMap ();

		for (Iterator i = edges.values ().iterator (); i.hasNext (); )
		{
			Edge e = (Edge) i.next ();
			pushOntoViaMap (vias, e, e.vertex[0].name);
			pushOntoViaMap (vias, e, e.vertex[1].name);
			pushOntoViaMap (vias, e, e.road);
			if (e.via != null)
			{
				for (Enumeration v = e.via.elements(); v.hasMoreElements() ;) 
				{
					pushOntoViaMap (vias, e, (String) v.nextElement ());
				}
			}
		}

		// write list of all journeys to the data file
		PrintWriter p = new PrintWriter (new FileOutputStream (filename));
		int index = 0;
		HashMap journeyIndex = new HashMap ();

		for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
		{
			Journey j = (Journey) i.next ();
			Edge e = j.findEdge (edges.values ().iterator ());
			if (e != null)
			{
				journeyIndex.put (j, new Integer (index++));
				p.print (j.getHandle ());
				p.print (":");
				p.print ("Journey from "+j.from.name+" to "+j.to.getUltimateDestination (e));
				if (j.getDetail ("en") != null)
				{
					p.print (" (");
					p.print (j.getDetail ("en"));
					p.print (")");
				}
				p.println ();
			}
		}
		p.println ();
	
		for (Iterator i = vias.keySet ().iterator (); i.hasNext (); )
		{
			String name = (String) i.next ();
			HashSet edgeSet = (HashSet) vias.get (name);
			String qualifier = "";
			int paren = name.indexOf ('(');
			boolean first = true;
			if (paren > -1)
			{
				qualifier = name.substring (paren + 1, name.length () - 1);
				name = name.substring (0, paren - 1);
			}
			p.print (name);
			p.print (":");
			p.print (qualifier);
			p.print (":");
			
			// now collect all journey indexes for this keyword
			// (all journeys that use an edge which claims this keyword)
			for (Iterator j = edgeSet.iterator (); j.hasNext (); )
			{
				Edge testEdge = (Edge) j.next ();
				for (Iterator k = journeys.values ().iterator (); k.hasNext (); )
				{
					Journey testJourney = (Journey) k.next ();
					if (testJourney.edgeMatches (testEdge))
					{
						if (!first) p.print (",");
						first = false;
						p.print (journeyIndex.get (testJourney));
					}
				}
			}
			p.println ();
		}
		p.close ();

	}

	public void createLocationIndex (String filename)
	throws IOException
	{
		Vector points = new Vector ();
		int index = 0;
		PrintWriter p = new PrintWriter (new FileOutputStream (filename));
		double minlong = Double.MAX_VALUE;
		double minlat = Double.MAX_VALUE;
		double maxlat = -Double.MAX_VALUE;

		// build index of all coordinate points
		for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
		{
			Journey j = (Journey) i.next ();
			Edge e = j.findEdge (edges.values ().iterator ());
			if (e != null)
			{
				p.print (j.getHandle ());
				p.print (":");
				p.print ("Journey from "+j.from.name+" to "+ j.to.getUltimateDestination (e));
				if (j.getDetail("en") != null)
				{
					p.print (" (");
					p.print (j.getDetail ("en"));
					p.print (")");
				}
				p.println ();
				Integer indexObject = new Integer (index++);
				for (int k = 0; k < j.photo.length; k++)
				{
					Waypoint wp = j.photo[k].waypoint;
					points.add (new Object[] { wp.coord, indexObject });
					if (wp.coord.x < minlong) minlong = wp.coord.x;
					if (wp.coord.y < minlat) minlat = wp.coord.y;
					if (wp.coord.y > maxlat) maxlat = wp.coord.y;
				}
			}
		}

		double latFactor = 111120;
		double medianLatitude = (maxlat + minlat) / 2;
		double longFactor = latFactor * Math.cos (Math.toRadians (medianLatitude));

		p.println ();
		p.print (minlong);
		p.print (":");
		p.print (maxlat);
		p.print (":");
		p.print (longFactor);
		p.print (":");
		p.println (latFactor);


		// sort
		Collections.sort (points, new Comparator () {
			public int compare (Object a, Object b) {
				double y1 = ((Point2D.Double) ((Object[])a)[0]).y;
				double y2 = ((Point2D.Double) ((Object[])b)[0]).y;
				return (y1>y2) ? -1 : (y1<y2) ? 1 : 0;
			}
		});

		// write
		for (Enumeration e = points.elements (); e.hasMoreElements (); )
		{
			Object[] array = (Object[]) e.nextElement ();
			Point2D.Double pt = (Point2D.Double) array[0];
			p.print ((int)((pt.x - minlong) * longFactor));
			p.print (":");
			p.print ((int)((maxlat - pt.y) * latFactor));
			p.print (":");
			double os[] = GridConverter.degreeToGrid (pt.y, pt.x);
			p.print ((int) (os[0] + 0.5));
			p.print (":");
			p.print ((int) (os[1] + 0.5));
			p.print (":");
			 
			p.println ((Integer) array[1]);
/*
			p.print (":");
			p.print (((Point2D.Double) array[0]).x); 
			p.print (":");
			p.println (((Point2D.Double) array[0]).y); 
*/
		}
		p.close ();

	}

	private void pushOntoViaMap (HashMap vias, Edge edge, String name)
	{
		if ((name == null) || name.equals (""))
		{
			return;
		}
		HashSet s = (HashSet) vias.get (name);
		if (s == null)
		{
			s = new HashSet ();
			vias.put (name, s);
		}
		s.add (edge);
	}


	public void createDotMaps (Map m, String mapDir)
	throws IOException
	{

		AffineTransform identity = new AffineTransform ();

		// finally, see what "specialized" map images (with dots painted into them)
		// are required and generate them. the step above will already have registered
		// the need for such images in the dotImageRequests array. it's usually one
		// per photo but may be less if the photos "share" a dot (either two images
		// of one journey falling onto one dot because of scaling, or images from
		// different journeys)

		for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
		{
			SizeSpec spec = (SizeSpec) e.nextElement ();
			if (spec.mapScale == 0) continue;

			MapInfo mi = (MapInfo) m.mapInfo.get (spec.name);
			for (int row = 0; row<mi.rows; row++)
			for (int col = 0; col<mi.columns; col++)
			{
				try
				{
					MagickImage partImage = new MagickImage (new ImageInfo (mi.partFilename[row][col]));
					for (Iterator i = mi.dotImageRequests[row][col].keySet ().iterator(); i.hasNext (); )
					{
						Point dot = (Point) i.next ();
						String name = mapDir + m.getHandle () + "/" + spec.getLetter () + "/" + (row+1) + "x" + (col+1) + "+" + dot.x + "x" + dot.y + ".gif";
						if (incremental && fileExists (name))
						{
							continue;
						}
						Point dotPos = new Point (dot.x - mi.partRect[row][col].x,
							dot.y - mi.partRect[row][col].y);
						int radius = (int) (6.0 * Math.sqrt ((double) spec.mapScale / 100.0));

						ImageInfo ii = new ImageInfo (name);
						DrawInfo di = new DrawInfo (ii);
						di.setPrimitive ("fill black circle "+dotPos.x+","+dotPos.y+" "+
							(dotPos.x+radius)+","+(dotPos.y+radius));
						MagickImage newMap = partImage.cloneImage (0,0,true);
						newMap.drawImage (di);
						di.setPrimitive ("fill red circle "+dotPos.x+","+dotPos.y+" "+
							(dotPos.x+radius-1)+","+(dotPos.y+radius-1));
						newMap.drawImage (di);
						newMap.setFileName (name);
						report ("    writing " + name);
						newMap.setNumberColors (spec.mapColors);
						newMap.writeImage (ii);
						newMap.destroyImage ();
						increaseMagickUsage ();
					}
					partImage.destroyImage ();
				}
				catch (MagickException me)
				{
					throw new IOException (me.getMessage ());
				}
			}
		}
	}

	public void processJourney (Journey j, Edge edge, String staticDir, String dynamicDir)
	throws IOException
	{

		NumberFormat nf = NumberFormat.getInstance ();
		nf.setMinimumIntegerDigits ((j.photo.length < 100) ? 2 : 3);
		nf.setParseIntegerOnly (true);
		HashMap writers = new HashMap ();
		AffineTransform transform = new AffineTransform ();
		StringBuffer mapSizeLine = new StringBuffer ();
		StringBuffer imageSizeLine = new StringBuffer ();
		
		for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
		{
			SizeSpec spec = (SizeSpec) e.nextElement ();
			if (spec.mapScale != 0)
			{
				if (mapSizeLine.length () > 0) mapSizeLine.append (':');
				mapSizeLine.append (spec.name);
			}
			if (spec.imageX != 0)
			{
				if (imageSizeLine.length () > 0) imageSizeLine.append (':');
				imageSizeLine.append (spec.name);
			}
		}

		// create directories if required; initialize the journey control files
		for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
		{
			SizeSpec spec = (SizeSpec) e.nextElement ();
			File path = new File (staticDir + spec.getLetter ());
			path.mkdirs ();

			if (spec.mapScale == 0) continue;

			path = new File (dynamicDir + spec.getLetter ());
			path.mkdirs ();
			report ("    writing " + dynamicDir + spec.getLetter () + "/control.txt");
			PrintWriter p = new PrintWriter (new FileOutputStream (dynamicDir +
				spec.getLetter () + "/control.txt"));
			writers.put (spec.name, p);

			// generate a line that reads like this:
			// Cove:Poolewe:Wester Ross (Scotland):B8057:64
			/* if (!j.from.name.equals (j.to.name)) */ p.print (j.from.name);
			p.print (":");
			p.print (j.to.getUltimateDestination (j.getEdge ()));
			p.print ("::"); // area code left empty
			if (edge.road != null) p.print (edge.road);
			p.print (":");
			p.println (j.photo.length);

			// generate a line like this:
			// 256:192:bla-blub:3:2:116:163:54,54,55:65,51
			// decodes as imagewidth:imageheight:mapname:maprows:mapcols:mapwidth:mapheight:
			//       rowheights:colwidths (for map segments; comma separated)
			p.print (spec.imageX);
			p.print (":");
			p.print (spec.imageY);
			p.print (":");
			p.print (edge.map.getHandle ());
			p.print (":");
			if (spec.mapScale == 0)
			{
				p.print ("0:0:0:0::");
			}
			else
			{
				MapInfo mi = (MapInfo) edge.map.mapInfo.get (spec.name);
				p.print (mi.width);
				p.print (":");
				p.print (mi.height);
				p.print (":");
				p.print (mi.rows);
				p.print (":");
				p.print (mi.columns);
				p.print (":");

				if ((mi.rows == 1) && (mi.columns == 1))
				{
					p.print (":");
				}
				else
				{
					for (int i=0; i<mi.rows; i++)
					{
						p.print (mi.partRect[i][0].height);
						if (i<mi.rows-1) p.print (",");
					}
					p.print (":");
					for (int i=0; i<mi.columns; i++)
					{
						p.print (mi.partRect[0][i].width);
						if (i<mi.columns-1) p.print (",");
					}
				}
			}

			p.println ();	

			// generate a line about the available sizes in ascending order like this:
			// tiny:small:medium:large:huge
			p.println (mapSizeLine);	
			p.println (imageSizeLine);	
			p.println ();
		}

		for (int i=0; i < j.photo.length; i++)
		{
			// this produces one jpg file for each photo and size class
			String fileName = nf.format (i+1) + ".jpg";

			// convert this photo's GPS coordinates into pixels 
			j.photo[i].waypoint.determineMapPixel (edge.map);
			/*
			Point2D.Double geo = new Point2D.Double (j.photo[i].waypoint.longitude, 
				j.photo[i].waypoint.latitude);
			Point pixel = new Point ();
			edge.map.mapGeoToPixel (geo, pixel);
			edge.map.applyCorrection (pixel);
			*/
			Point pixel = j.photo[i].waypoint.mapPixel;

			MagickImage srcImage = null;

			if (!mapsOnly)
			{
				if (incremental)
				{
					// we may not have to load the image at all - only load if at least
					// one of the target images diesn't exist
					for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
					{
						String name = staticDir + ((SizeSpec) e.nextElement ()).getLetter () + "/" + fileName;
						if (! fileExists (name))
						{
							srcImage = fetchAndPreprocessImage (j.photo[i]);
							break;
						}
					}
				}
				else
				{
					srcImage = fetchAndPreprocessImage (j.photo[i]);
				}
			}

			for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
			{
				SizeSpec spec = (SizeSpec) e.nextElement ();

				if (!mapsOnly)
				{
					String name = staticDir + spec.getLetter () + "/" + fileName;
					if (!(incremental && fileExists (name)))
					{
						report ("      writing " + name);
						try
						{
							MagickImage newImage = srcImage.scaleImage (spec.imageX, spec.imageY);
							ImageInfo newImageInfo = new ImageInfo (name);
							newImage.setFileName (name);
							newImageInfo.setQuality (spec.imageQuality);
							newImage.writeImage (newImageInfo);
							newImage.destroyImage ();
							increaseMagickUsage ();
						}
						catch (MagickException m)
						{
							throw new IOException (m.getMessage ());
						}
					}
				}

				if (spec.mapScale == 0) continue;

				MapInfo mi = (MapInfo) edge.map.mapInfo.get (spec.name);
				Point scaledPos = new Point (pixel.x * spec.mapScale / 100, 
					pixel.y * spec.mapScale / 100);
				int radius = (int) (8.0 * Math.sqrt ((double) spec.mapScale / 100.0));
				Rectangle pointEnclosingRect = new Rectangle (scaledPos.x - radius, scaledPos.y - radius,
					2*radius, 2*radius);
				PrintWriter pw = (PrintWriter) writers.get (spec.name);
				StringBuffer mapSegName = new StringBuffer ();

				// find related map segment 
				// boolean preferred = true;
				for (int row=0; row<mi.rows; row++)
				for (int col=0; col<mi.columns; col++)
				{
					if (mi.partRect[row][col].intersects (pointEnclosingRect))
					{
						mi.dotImageRequests[row][col].put (scaledPos, null);
						if (mapSegName.length () != 0) mapSegName.append (",");
						mapSegName.append (row+1);
						mapSegName.append ("x");
						mapSegName.append (col+1);
					}
				}

				// stuff info about this image into control file
				pw.print (fileName);
				pw.print (":");
				pw.print (scaledPos.x);
				pw.print (":");
				pw.print (scaledPos.y);
				pw.print (":");
				pw.print (mapSegName);
				pw.print (":");
				pw.print (j.photo[i].waypoint.prettyLatitude ());
				pw.print (":");
				pw.print (j.photo[i].waypoint.prettyLongitude ());
				pw.print (":");
				pw.print ((int) (j.photo[i].waypoint.time / 1000));
				pw.print (":");
				pw.print (j.photo[i].waypoint.speed);
				pw.print (":");
				pw.println (j.photo[i].waypoint.heading);
				
			}

			if (srcImage != null) srcImage.destroyImage ();
		}

		// where we are
		Vertex ici = j.to;
		String leg = edge.connector[(edge.vertex[0] == ici) ? 0 : 1];
		int fromConnector = ici.findConnectorIndex (leg);
		if (fromConnector == -1)
		{
			throw new IOException ("Edge points to non-existent connector '"+leg+"' at vertex '"+ici.name+"'");
		}

		// finish off the print writers by putting an empty line and information about
		// all connecting journeys at the end.

		for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
		{
			SizeSpec spec = (SizeSpec) e.nextElement ();
			if (spec.mapScale == 0) continue;

			PrintWriter pw = (PrintWriter) writers.get (spec.name);
			
			pw.println ();


			pw.println (ici.getDetail("en")); // FIXME write all languages
			
			// one line for each connection.
			// format like this:
			// o:270:A832:Aultbea:poolewe-aultbea:out:april
			// type (o=option d=default b=back):turn (degrees):road name:destination:edge handle:
			// journey id

			for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
			{
				Journey jj = (Journey) i.next ();
				if (jj.from != ici) continue;
				Edge ee = jj.findEdge (edges.values ().iterator ());
				String theConnector = ee.connector [(jj.from == ee.vertex[0]) ? 0 : 1];

				int onDir = -1;
				int onConnector;

				onConnector = ici.findConnectorIndex (theConnector);
				if (onConnector == -1)
				{
					throw new IOException ("Edge points to non-existent connector '" +
						theConnector + "' at vertex '"+jj.from.name+"'");
				}
				onDir = ici.direction[onConnector] - ici.direction[fromConnector] + 540;
				if (onDir > 360) onDir -= 360;
				if (onDir > 360) onDir -= 360;
				pw.print (ici.isPassthrough (fromConnector, onConnector) ? "d" : (fromConnector == onConnector) ? "b" : "o");
				pw.print (":");
				pw.print (jj.priority);
				pw.print (":");
				if (jj.getDetail ("en") != null) pw.print (jj.getDetail ("en"));
				pw.print (":");
				pw.print (onDir);
				pw.print (":");
				pw.print (ee.road);
				pw.print (":");
/*
				if (!jj.to.name.equals (ici.name))
				{
*/
					pw.print (jj.to.getUltimateDestination (ee)); /* signpost? */
/*
				}
*/
				pw.print (":");
				pw.print (ee.getHandle ());
				pw.print (":");
				pw.println (jj.getHandle ());
			}

			if (ici.hasDescription ())
			{
				pw.print (ici.getBaseName ());
				pw.println (".html");
			}
			pw.close ();
		}
		
	}

	public MagickImage fetchAndPreprocessImage (Photo photo)
	throws IOException 
	{
		report ("    reading " + photo.filename);

		try
		{
			ImageInfo info = new ImageInfo (photo.filename);
			MagickImage orig = new MagickImage (info);
			
			if (quickAndDirty)
			{
				return orig;
			}

			if (photo.getGamma() != 0)
			{
				orig.gammaImage (Double.toString ((100.0 + photo.getGamma ())/100.0));
			}
			if (photo.getBrightness() != 0)
			{
				orig.modulateImage (Integer.toString (100 + photo.getBrightness ()));
			}
			return orig;
		}
		catch (MagickException m)
		{
			throw new IOException (m.getMessage ());
		}
		
	}


	public void processMap (Map map, String mapDir)
	throws IOException
	{
		map.mapInfo = new HashMap ();

		ImageInfo ii = null;
		MagickImage mapImage = null;
		
		for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
		{
			SizeSpec spec = (SizeSpec) e.nextElement ();
			if (spec.mapScale == 0) continue;
			if (vertexOnly && !spec.name.equals ("medium")) continue;

			String dir = mapDir + map.getHandle () + "/" + spec.getLetter () + "/";
			new File (dir).mkdirs ();

			MapInfo mi = new MapInfo ();

			mi.width = map.size.width * spec.mapScale / 100;
			mi.height = map.size.height * spec.mapScale / 100;

			MagickImage sizedMap = null;

			int rows[] = new int[map.rowDivider == null ? 2 : map.rowDivider.length + 2];
			int cols[] = new int[map.colDivider == null ? 2 : map.colDivider.length + 2];
			rows[0] = 0;
			cols[0] = 0;
			rows[rows.length-1] = mi.height;
			cols[cols.length-1] = mi.width;
			for (int i=0; i<rows.length-2; i++) rows[i+1]=map.rowDivider[i] * spec.mapScale / 100;
			for (int i=0; i<cols.length-2; i++) cols[i+1]=map.colDivider[i] * spec.mapScale / 100;

			mi.rows = rows.length - 1;
			mi.columns = cols.length - 1;

			mi.partRect = new Rectangle[mi.rows][mi.columns];
			mi.partFilename = new String[mi.rows][mi.columns];
			mi.dotImageRequests = new HashMap[mi.rows][mi.columns];
			for (int row = 0; row < mi.rows; row++) 
			for (int col = 0; col < mi.columns; col++)
			{
				mi.dotImageRequests[row][col] = new HashMap ();
				mi.partRect[row][col] = new Rectangle (cols[col], rows[row], 
					cols[col+1]-cols[col], rows[row+1]-rows[row]);

					
				if (vertexOnly) continue;

				String name = dir + (row+1) + "x" + (col+1) + ".gif";

				try
				{
					mi.partFilename[row][col] = name;
					if (!(incremental && fileExists (name)))
					{
						// lazily create the scaled map image
						if (sizedMap == null)
						{
							// lazily load the large source map image
							if (mapImage == null)
							{
								ii = new ImageInfo (map.filename);
								report ("    loading " + map.filename);
								mapImage = new MagickImage (ii);
							}

							if (spec.mapScale < 40)
							{
								MagickImage tmpImage = mapImage.scaleImage (mi.width, mi.height);
								sizedMap = tmpImage.sharpenImage (10, 2);
								tmpImage.destroyImage ();
							} 
							else if (spec.mapScale < 50)
							{
								MagickImage tmpImage = mapImage.scaleImage (mi.width, mi.height);
								sizedMap = tmpImage.sharpenImage (8, 1);
								tmpImage.destroyImage ();
							}
							else
							{
								sizedMap = mapImage.scaleImage (mi.width, mi.height);
							}
						}
						MagickImage partImage = sizedMap.cropImage (mi.partRect[row][col]);

						if (partImage != null)
						{
							partImage.setNumberColors (spec.mapColors);
							report ("      creating " + name + ", " + mi.partRect[row][col].width + "x" + 
								mi.partRect[row][col].height + "+" + mi.partRect[row][col].x + "+" +
								mi.partRect[row][col].y);
							partImage.setFileName (name);
							partImage.writeImage (ii);
							increaseMagickUsage ();
							partImage.destroyImage ();
						}
					}
				}
				catch (MagickException m) 
				{
					throw new IOException (m.getMessage ());
				}

			}
					
			if (sizedMap != null) sizedMap.destroyImage ();
			map.mapInfo.put (spec.name, mi);
		}
		if (mapImage != null) mapImage.destroyImage ();
		
	}

	private boolean fileExists (String name)
	{
		boolean rv = false;
		try
		{
			rv = new File (name).exists ();
		}
		catch (Exception e)
		{
			// ignore;
		}
		return rv;
	}

	public void setMapsOnly (boolean mo)	
	{
		mapsOnly = mo;
	}

	public boolean getMapsOnly ()
	{
		return mapsOnly;
	}
	public void setVertexOnly (boolean vo)	
	{
		vertexOnly = vo;
	}

	public boolean getVertexOnly ()
	{
		return vertexOnly;
	}


	public void setIncremental (boolean in)
	{
		incremental = in;
	}
	public boolean getIncremental ()
	{
		return incremental;
	}

	////////////////////////////////
	// @@ Static "main" method
	////////////////////////////////

	/**
	 * Main method. 
	 * Parses arguments and calls appropiate methods on compiler object;
	 * finally asks compiler object to produce output.
	 */
	public static void main(String argv[])
	{

		Compiler theCompiler = new Compiler ();
		Getopt g = new Getopt ("compiler", argv, ":hvd:qmfx");
		int c;

		g.setOpterr (false);

		while ((c = g.getopt ()) != -1)
		{
			switch (c)
			{
			  case 'q':
				theCompiler.setQuickAndDirty (true);
				break;
			  case 'd':
				theCompiler.setDestinationPath (g.getOptarg ());
				break;
			  case 'v':
				theCompiler.setReportingStream (System.out);
				break;
			  case 'f':
				theCompiler.setIncremental (false);
				break;
			  case 'm':
				theCompiler.setMapsOnly (true);
				break;
			  case 'x':
				theCompiler.setVertexOnly (true);
				break;
			  default:
				if (c == ':') 
				{
					System.out.println ("-" + (char)g.getOptopt () + " requires an argument\n");
				}
				else if (c != 'h')
				{
					System.out.println ("Invalid option: -" + (char) g.getOptopt () + "\n");
				}
				System.out.println ("Usage:");
				System.out.print ("java " + theCompiler.getClass ().getName ());
				System.out.println (" [option]... [inputfile]...");
				System.out.println ("   -h             displays this help");
				System.out.println ("   -v             verbose progress report");
				System.out.println ("   -q             quick operation - no brightness/contrast oeprations");
				System.out.println ("   -d path        output destination path");
				System.out.println ("   -m             map processing only");
				System.out.println ("   -f             re-create files even if they exist");
				return;
			}
		}

		for (int i = g.getOptind(); i < argv.length; i++)
		{
			try
			{
				theCompiler.readFile (argv[i]);
			}
			catch (Exception ex)
			{
				System.err.println ("Error while processing file '" + argv[i] + "':");
				System.err.println ("   " + ex.getClass () + ": " + ex.getMessage ());
				if (theCompiler.getReportingStream() != null) ex.printStackTrace ();
			}
		}

		try
		{
			theCompiler.produceOutput ();
		}
		catch (Exception ex)
		{
			System.err.println ("Error while producing output:");
			System.err.println ("   " + ex.getClass () + ": " + ex.getMessage ());
			if (theCompiler.getReportingStream() != null) ex.printStackTrace ();
		}

		System.exit (0);
	}

	private void increaseMagickUsage ()
	{
		magickUsage++;
		if ((magickUsage % 25) == 0)
		{
			// System.runFinalization();
			// System.gc();
		}
	}

	private String headingToWord (int heading, String lang)
	{
		if (lang.equals ("en"))
		{
			if (heading < 23)
				return "north";
			else if (heading < 68)
				return ("northeast");
			else if (heading < 113)
				return ("east");
			else if (heading < 158)
				return ("southeast");
			else if (heading < 203)
				return ("south");
			else if (heading < 248)
				return ("southwest");
			else if (heading < 293)
				return ("west");
			else if (heading < 338)
				return ("northwest");
			else
				return "north";
		}

		if (lang.equals ("de"))
		{
			if (heading < 23)
				return "Norden";
			else if (heading < 68)
				return ("Nordosten");
			else if (heading < 113)
				return ("Osten");
			else if (heading < 158)
				return ("Sdosten");
			else if (heading < 203)
				return ("Sden");
			else if (heading < 248)
				return ("Sdwesten");
			else if (heading < 293)
				return ("Westen");
			else if (heading < 338)
				return ("Nordwesten");
			else
				return "Norden";
		}

		return null;
	}

	////////////////////////////////
	// @@ Class SizeSpec
	////////////////////////////////

	private class SizeSpec 
	{
		public int imageX;
		public int imageY;
		public int imageQuality;
		public int mapScale;
		public int mapColors;
		public String name;

		public SizeSpec (int a, int b, int c, int d, int e, String f)
		{
			imageX = a; imageY = b; imageQuality = c; mapScale = d; mapColors = e; name = f;
		}

		/**
		 * Returns unique code letter for this size code. (This implementation simply
		 * returns the first letter.) Has to match up with CGI code.
		 */
		private String getLetter ()
		{
			return name.substring (0,1);
		}
	}

	private class MapInfo
	{
		public Rectangle[][] partRect;
		public String[][] partFilename;
		public HashMap[][] dotImageRequests;
		public int rows;
		public int columns;
		public int height;
		public int width;
	}

}
