
package org.remote.roadhog;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;

import org.w3c.dom.*;

import org.apache.xerces.parsers.DOMParser;

class Edge extends XMLObject
{
	public String road = "";
	public String alternative = null;
	public int length;
	public Vertex[] vertex = new Vertex[2];
	public String[] connector = new String[2];
	public Vector via = new Vector ();
	public Vector flags = new Vector ();
	public Map map;

	/**	
	 * Creates an empty Edge object.
	 */
	public Edge ()
	{
		/* empty */
	}

	/**
	 * Creates a new Edge object from the given XML node.
	 */
	public Edge (Node node, XMLObjectLookup lookup)
	throws ParseException
	{
        NodeList sublist;
        Element edge = (Element) node;

        sublist = edge.getElementsByTagName ("road");
        if (sublist.getLength () > 0) road = getTextChild (sublist.item(0));

        sublist = edge.getElementsByTagName ("length");
        if (sublist.getLength () > 0) length = Integer.parseInt (getTextChild (sublist.item(0)));

        sublist = edge.getElementsByTagName ("alternative");
        if (sublist.getLength () > 0) alternative = getTextChild (sublist.item(0));

        sublist = edge.getElementsByTagName ("vertex-id");
		for (int i=0; i<2; i++)
		{	
			Object vr[] = extractVertexReference (sublist.item (i), lookup);
			vertex[i] = (Vertex) vr[0];
			connector[i] = (String) vr[1];
		}

        sublist = edge.getElementsByTagName ("via");
		for (int i=0; i<sublist.getLength (); i++) via.add (getTextChild (sublist.item (i)));
		
        sublist = edge.getElementsByTagName ("flag");
		for (int i=0; i<sublist.getLength (); i++) flags.add (getTextChild (sublist.item (i)));

        sublist = edge.getElementsByTagName ("map-id");
        String mapid =  getTextChild (sublist.item(0));

        map = (Map) lookup.findObject (Map.class, mapid);
		if (map == null)
        {
            throw new ParseException ("Unresolved reference to Map id='" + mapid + "'");
        }
	}

	/**
	 * Helper for the XML parsing constructor
	 */
	private Object[] extractVertexReference (Node n, XMLObjectLookup lookup)
	throws ParseException
	{
        NodeList sublist;
        Element vertEl =  (Element) n;
        sublist = vertEl.getElementsByTagName ("id");
		String vertId = getTextChild (sublist.item (0));
        Vertex vert = (Vertex) lookup.findObject (Vertex.class, vertId);
		if (vert == null)
        {
            throw new ParseException ("Unresolved reference to Vertex id='" + vertId + "'");
        }
        sublist = vertEl.getElementsByTagName ("connector-id");
		String conn = getTextChild (sublist.item (0));
		for (int i=0; i < vert.connector.length; i++)
		{
			if (vert.connector[i].equals (conn))
			{
				vert.edge[i] = this;
				return new Object[] { vert, conn };
			}
		}
		throw new ParseException ("Bad reference to connector '" + conn + 
				"' in vertex '" + vertId + "'");
	}

	public void toXML (PrintStream o, String indent)
	{
		o.println (indent + "<edge" + getOutputAttributes () + ">");
		if (road != null) o.println (indent + "   <road>" + stringToXML (road) + "</road>");
		if (length != 0) o.println (indent + "   <length units=\"miles\">" +  length + "</length>");
		if (alternative != null) o.println (indent + "   <alternative>"+alternative + "</alternative>");
		o.println (indent + "   <vertex-id>");
		o.println (indent + "      <id>" + stringToXML (vertex[0].id) + "</id>");
		o.println (indent + "      <connector-id>" + stringToXML (connector[0]) + "</connector-id>");
		o.println (indent + "   </vertex-id>");
		o.println (indent + "   <vertex-id>");
		o.println (indent + "      <id>" + stringToXML (vertex[1].id) + "</id>");
		o.println (indent + "      <connector-id>" + stringToXML (connector[1]) + "</connector-id>");
		o.println (indent + "   </vertex-id>");
		if (map != null) o.println (indent + "   <map-id>" + map.id + "</map-id>");
		
		for (Enumeration e = via.elements (); e.hasMoreElements (); )
			o.println (indent + "   <via>" + stringToXML ((String) e.nextElement ()) + "</via>");
		for (Enumeration f = flags.elements (); f.hasMoreElements (); )
			o.println (indent + "   <flag>" + stringToXML ((String) f.nextElement ()) + "</flag>");

		o.println (indent + "</edge>");
	}

	public String getLongHandle ()
	{
		StringBuffer handle;
		if (vertex[0].id.compareTo (vertex[1].id) > 0)
		{
			handle = new StringBuffer (vertex[1].id + "-" + vertex[0].id);
		}
		else
		{
			handle = new StringBuffer (vertex[0].id + "-" + vertex[1].id);
		}
		if (alternative != null) 
		{
			handle.append ("-");
			handle.append (alternative);
		}

		for (int i=0; i<handle.length (); i++)
		{
			if (handle.charAt (i) == ' ') handle.deleteCharAt (i--);
		}
		return (handle.toString ().toLowerCase ());
	}

	public String getHandle ()
	{	
		return getLongHandle ();
	}

	public String getShortHandle ()
	{
		MD5 m = new MD5 ();
		m.update (getLongHandle ().getBytes ());
		byte[] bytes = new byte[16];
		m.md5final (bytes);
		return dumpBytes (bytes).substring (0, 4);
	}

}
