/**
 * 
 * RoadHog JourneyMaker
 * v0.9
 * (c) 2001 Frederik Ramm 
 * Released under GPL
 *
 * I know this needs documentation. 
 * See http://www.remote.org/frederik/roadhog/software/ for details.
 */
package org.remote.roadhog;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.border.*;

import org.w3c.dom.*;
import org.apache.xerces.parsers.DOMParser;
import com.skynamics.lib.swing.SimpleGridBag;

import magick.*;

public class JourneyMaker extends Gui
{

	private JMenuItem editCut;
	private JMenuItem select;
	private JMenuItem deselect;
	private JMenuItem loadPhotos;
	private JMenuItem photoLocation;
	private JMenuItem clearHacks;

	private Edge currentEdge = new Edge ();
	private Journey currentJourney = new Journey ();

	private Waypoint[] waypoint;
	private JTable waypointTable;
	private WaypointTableModel waypointTableModel;

	private JTable photoTable;
	private PhotoTableModel photoTableModel;

	private Vertex[] vertex;
	private JTable vertexTable;
	private VertexTableModel vertexTableModel;

	private EdgeEditor edgeEdit; 

	private MapWithTrack mapWithTrack;
	private ImagePanel imagePanel;
	private int mode = 0;
	private final static int MODE_TRACK = 0;
	private final static int MODE_PHOTO = 1;
	private final static int MODE_VERTEX = 2;
	private final static int MODE_EDGE = 3;

	private double minLat = 0;
	private double maxLat = 0;
	private double minLong = 0;
	private double maxLong = 0;

	private JLabel coordLabel;

	private final static int dotSize = 4; // radius
	private DOMParser parser = new DOMParser();

	private boolean debug = true;
	private final static boolean LINES_ONLY_IF_SELECTED=true;


	public JourneyMaker ()
	{

		super ("RoadHog JourneyMaker v0.9");

		JMenuItem menuItem;
		
		currentEdge.map = new Map ();

		setSize (1000, 600);

		setJMenuBar (new JMenuBar ());
		JMenu fileMenu = new JMenu ("File");
		getJMenuBar ().add (fileMenu);


		fileMenu.add (menuItem = new JMenuItem ("Load Vertices..."));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { loadVertices (); }
		});

		fileMenu.add (menuItem = new JMenuItem ("Load XML..."));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { loadXML (); }
		});

		fileMenu.add (menuItem = new JMenuItem ("Save XML..."));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { saveXML (); }
		});

		fileMenu.addSeparator ();

		fileMenu.add (menuItem = new JMenuItem ("Import Track..."));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { loadTrack (); }
		});

		fileMenu.add (loadPhotos = new JMenuItem ("Import Photos..."));
		loadPhotos.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { loadPhotos (); }
		});

		fileMenu.addSeparator ();

		fileMenu.add (menuItem = new JMenuItem ("Quit"));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { quit (); }
		});

		JMenu editMenu = new JMenu ("Edit");
		getJMenuBar ().add (editMenu);

		editMenu.add (editCut = new JMenuItem ("Cut"));
		editCut.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { editCut (); }
		});
		editMenu.add (select = new JMenuItem ("Select"));
		select.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { select (true); }
		});
		select.setAccelerator (KeyStroke.getKeyStroke ("control S"));
		editMenu.add (deselect = new JMenuItem ("Deselect"));
		deselect.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { select (false); }
		});
		deselect.setAccelerator (KeyStroke.getKeyStroke ("control D"));
		photosSelected (false);

		editMenu.add (photoLocation = new JMenuItem ("Time correction..."));
		photoLocation.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { photoLocationMapping (); }
		});
		photoLocation.setEnabled (false);
		editMenu.add (clearHacks = new JMenuItem ("Clear all hacks"));
		clearHacks.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { clearHacks (); }
		});

		final JTabbedPane tabPane = new JTabbedPane ();
		tabPane.addTab ("GPS Track", null, 
			new JScrollPane (waypointTable = new JTable (waypointTableModel = new WaypointTableModel ())), null);
		waypointTable.getSelectionModel().addListSelectionListener (new ListSelectionListener () {	
			public void valueChanged (ListSelectionEvent e)
			{
				int index = waypointTable.getSelectedRow ();
				if (index >= 0)
				{
					mapWithTrack.repaint ();
				}
			}
		});
		tabPane.addTab ("Photos", null, 
			new JSplitPane (JSplitPane.VERTICAL_SPLIT, 
			new JScrollPane (photoTable = new JTable (photoTableModel = new PhotoTableModel ())), 
				imagePanel = new ImagePanel ()), null);
		photoTable.getSelectionModel().addListSelectionListener (new ListSelectionListener () {	
			public void valueChanged (ListSelectionEvent e)
			{
				int index = photoTable.getSelectedRow ();
				if (index >= 0)
				{
					imagePanel.setPhoto (currentJourney.photo[index]);
				}
				photosSelected (mode == MODE_PHOTO && index >= 0);
				mapWithTrack.repaint ();
			}
		});
		photoTableModel.setColumnWidths (photoTable);
		photoTableModel.setColumnRenderers (photoTable);

/*
		ActionListener edgeListener = new ActionListener () {
			public void actionPerformed (ActionEvent a) {
				edgeEditCommand (a.getActionCommand ());
			}
		};
*/

		tabPane.addTab ("Vertices", null, 
			new JScrollPane (vertexTable = new JTable (vertexTableModel = new VertexTableModel ())), null);
		tabPane.addTab ("Edge Data", null, 
			new JScrollPane (edgeEdit = new EdgeEditor (currentEdge, currentJourney, vertex)), null);

		tabPane.addChangeListener (new ChangeListener () {
			public void stateChanged (ChangeEvent e)
			{
				switch (tabPane.getSelectedIndex ())
				{
				  case 0:
					mode = MODE_TRACK;
					photosSelected (false);
					break;
				  case 1:
					mode = MODE_PHOTO;
					photosSelected (tabPane.getSelectedIndex () == 1 && 
						photoTable.getSelectedRow () > -1);
					break;
				  case 2:
					mode = MODE_VERTEX;
					break;
				  case 3:
					mode = MODE_EDGE;
					break;
				}
				mapWithTrack.repaint ();
			}
		});

		JPanel mapPanel = new JPanel ();
		JPanel buttonPanel = new JPanel ();
		JPanel southPanel = new JPanel ();
		mapPanel.setLayout (new BorderLayout ());
		mapPanel.add (new JScrollPane (mapWithTrack = new MapWithTrack ()), BorderLayout.CENTER);
		mapPanel.add (southPanel, BorderLayout.SOUTH);
		southPanel.setLayout (new BorderLayout ());
		southPanel.add (buttonPanel, BorderLayout.CENTER);
		southPanel.add (coordLabel = new JLabel (" "), BorderLayout.NORTH);
		coordLabel.setBorder (new EmptyBorder (5, 5, 0, 5));

		final JSlider s = new JSlider (-2, 2, 0);
		s.setSnapToTicks (true);
		s.setPaintTicks (true);
		s.setMajorTickSpacing (1);
		buttonPanel.add (s);
		s.addChangeListener (new ChangeListener () {
			public void stateChanged (ChangeEvent e) {
				switch (s.getValue ())
				{
				  case -2:
					mapWithTrack.setScale (25); break;
				  case -1:
					mapWithTrack.setScale (50); break;
				  case 0:
					mapWithTrack.setScale (100); break;
				  case 1:
					mapWithTrack.setScale (200); break;
				  case 2:
					mapWithTrack.setScale (400); break;
				}
			}
		});
				
		JSplitPane splitPane = new JSplitPane (JSplitPane.HORIZONTAL_SPLIT, 
			mapPanel, tabPane);
		
		Container main = getContentPane ();
        main.setLayout (new BorderLayout ());
		main.add (splitPane, BorderLayout.CENTER);
		

	}

	private void clearHacks ()
	{
		if ((currentJourney != null) && (currentJourney.photo != null))
		{
			for (int i=0; i<currentJourney.photo.length; i++)
			{
				currentJourney.photo[i].waypoint.hack = null;
				currentJourney.photo[i].waypoint.determineMapPixel (currentEdge.map);
			}
			mapWithTrack.repaint ();
		}
	}



	private void photosSelected (boolean selected)
	{
		editCut.setEnabled (selected);
		select.setEnabled (selected);
		deselect.setEnabled (selected);
	}

	private void select (boolean select)
	{

		int selRow = photoTable.getSelectedRow ();
		int selCount = photoTable.getSelectedRowCount ();

		for (int i = selRow; i<selRow + selCount; i++)
		{
			currentJourney.photo[i].selected = select;
		}
		mapWithTrack.repaint ();
		photoTableModel.refresh ();
		
	}
	
	public void loadXML ()
	{
		loadXML (null);	
	}
	
	public void loadXML (File[] filelist)
	{
		XMLLoader loader = new XMLLoader (this, filelist);
		if (currentEdge.map != null) loader.addMap (currentEdge.map);
		boolean updateMapPixels = false;

		if (loader.hasError ())
			return;

		loader.setVisible (true);
		Edge e = loader.getSelectedEdge ();
		if (e != null)
		{
			currentEdge = e;
			// update what?
			edgeEdit.setEdge (currentEdge);
		}
		Map m = loader.getSelectedMap ();
		if (m != null)
		{
			currentEdge.map = m;
			edgeEdit.updateMap ();
			if (waypoint != null)
			{
				for (int i=0; i<waypoint.length; i++) waypoint[i].determineMapPixel (m);
			}
			updateMapPixels = true;
			// FIXME delete those waypoints that the map says are to be deleted
			if (currentJourney.photo != null)
			{
				for (int i=0; i<currentJourney.photo.length; i++) 
					currentJourney.photo[i].waypoint.determineMapPixel (m);
			}
			if (!mapWithTrack.setMap (m))
			{
				errorMessage ("Cannot load map file "+m.filename, null);
			}
			// resetTransformation ();
			coordLabel.invalidate ();
		}
		Journey j = loader.getSelectedJourney ();
		if (j != null)
		{
			currentJourney = j;
			waypoint = null;
			updateMapPixels = true;
			if (j.getWaypoints () != null)
			{
				Vector w = j.getWaypoints ();
				Point p = new Point ();

				// delete those waypoints which the map identifies as such
				if (currentEdge.map != null) for (int i=0; i<w.size (); i++)
				{
					Waypoint wp = (Waypoint) w.elementAt (i);
					currentEdge.map.mapGeoToPixel (wp.coord, p);
					if (currentEdge.map.isPixelDeleted (p))
					{
						w.removeElementAt (i);
					}
				}
				waypoint = new Waypoint[j.getWaypoints ().size ()];
				j.getWaypoints ().copyInto (waypoint);
				waypointTableModel.refresh ();
			}

			edgeEdit.setJourney (j);
			photoTableModel.refresh ();
			mapWithTrack.repaint ();
			loadThumbnails ();
			// TODO init the waypoints!
			// update..
		}

		if (updateMapPixels && (currentEdge.map != null))
		{
			if (waypoint != null) for (int i=0; i<waypoint.length; i++)
			{
				waypoint[i].determineMapPixel (currentEdge.map);
			}
			if (currentJourney.photo != null) for (int i=0; i<currentJourney.photo.length; i++)
			{
				currentJourney.photo[i].waypoint.determineMapPixel (currentEdge.map);
			}
		}
	}

	public void saveXML ()
	{
		File saveFile = fileSelect ("XML files", new String[] { "xml", "XML" }, true);
		if (saveFile == null) return;
		try 
		{
			PrintStream o = new PrintStream (new FileOutputStream (saveFile));
			o.println("<roaddb>");

			// save the edge
			try
			{
				ByteArrayOutputStream b = new ByteArrayOutputStream ();
				currentEdge.toXML (new PrintStream (b), "   ");
				o.print (b.toString ());
				o.println("");
			}
			catch (Exception e)
			{
				//ignore
			}

			// prepare the journey by adding all the relevant waypoints
			currentJourney.resetWaypoints ();
			int firstSelected = -1, lastSelected = -1;
			if (currentJourney.photo != null) for (int i=0; i < currentJourney.photo.length; i++)
			{
				if (currentJourney.photo[i].selected)
				{
					if (firstSelected == -1)	
					{
						firstSelected = i;
					}
					else
					{
						lastSelected = i;
					}
				}
			}

			if ((waypoint != null) && ((firstSelected > -1) && (lastSelected > -1)))
			{
				long earliestTime = currentJourney.photo[firstSelected].waypoint.time - 100000;
				long latestTime = currentJourney.photo[lastSelected].waypoint.time + 100000;
				for (int i=0; i<waypoint.length; i++)
				{
					if ((waypoint[i].time > earliestTime) &&
						(waypoint[i].time < latestTime))
					{
						currentJourney.addWaypoint (waypoint[i]);
					}
				}
			}

			try
			{
				ByteArrayOutputStream b = new ByteArrayOutputStream ();
				currentJourney.toXML (new PrintStream (b), "   ");
				o.print (b.toString ());
			}
			catch (Exception e)
			{
				//ignore
e.printStackTrace ();
			}
			o.println("</roaddb>");
			if (o.checkError ())
				throw new IOException (null);
			o.close ();
		}
		catch (IOException ex)
		{
			errorMessage ("Error saving " + saveFile.getName(), ex);
		}
	}

	public void loadVertices ()
	{
		File vertFile = fileSelect ("XML files", new String[] { "xml", "XML" });
		if (vertFile == null) return;
		loadVertices (vertFile);
	}

	public void loadVertices (File vertFile)
	{

		try		
		{
			parser.parse (vertFile.getCanonicalPath ());
		}
		catch (Exception e)
		{
			errorMessage ("Cannot open or parse file.", e);
			return;
		}
			
		NodeList vl = parser.getDocument ().getElementsByTagName ("vertex");
		int length = vl.getLength ();

		if (length > 0) 
		{
			vertex = new Vertex[length];
			for (int i=0; i<length; i++)
			{
				try
				{
					vertex[i] = new Vertex (vl.item (i));
				}
				catch (ParseException ex)
				{
					errorMessage ("Format error in file.", ex);
					vertex = null;
					return;
				}
			}
			System.out.println (length + " vertices read from " + vertFile);

			// sort by name
			Arrays.sort (vertex, new Comparator () {
				public int compare (Object a, Object b) {
					return ((Vertex)a).id.compareTo (((Vertex)b).id);
				}
			});

			vertexTableModel.refresh ();
			edgeEdit.setVertexArray (vertex);
		}
		else
		{
			errorMessage ("File does not contain vertex data.", null);
			return;
		}
	}

	public void loadTrack ()
	{
		File trackFile = fileSelect ();
		if (trackFile == null) return;
		loadTrack (trackFile);
	}

	public void loadTrack (File trackFile)
	{

		BufferedReader reader;
		Vector wplist = new Vector ();

		try
		{
			reader = new BufferedReader (new FileReader (trackFile));
		}
		catch (Exception e)
		{
			errorMessage ("Cannot open file.", e);
			return;
		}

		String line;
		int ok = 0;
		int err = 0;

		try {
			while ((line = reader.readLine ()) != null)
			{
				try
				{
					Waypoint w = new Waypoint (line);
					wplist.add (w);
					if (currentEdge.map != null) w.determineMapPixel (currentEdge.map);
					ok++;
				}
				catch (ParseException e)
				{
					err++;
				}
	
			}
	
			reader.close ();
		}
		catch (IOException e)
		{
			errorMessage ("Error reading file.", e);
			return;
		}

		if (err != 0)
		{
			StringBuffer msg = new StringBuffer ();
			msg.append (ok);
			msg.append (" lines parsed ok; ");
			msg.append (err);
			msg.append (" lines with errors");
			JOptionPane.showMessageDialog(this.getContentPane (), msg.toString(), 
				"Parse result", JOptionPane.INFORMATION_MESSAGE);
		}

		// System.out.println (ok + " waypoints read from " + trackFile);

		// convert into array and sort by time
		waypoint = new Waypoint[wplist.size ()];
		wplist.copyInto (waypoint);
		Arrays.sort (waypoint, new Comparator () {
			public int compare (Object a, Object b) { 
				return (((Waypoint)a).time < ((Waypoint)b).time) ? -1 :
				       (((Waypoint)a).time > ((Waypoint)b).time) ? 1 : 0;
			}
		});
		
		calculateWaypointDerivedValues ();
		// resetTransformation ();
		mapWithTrack.repaint ();
		waypointTableModel.refresh ();
		edgeEdit.updateJourney ();

	}

	/**
	 * Sorts waypoint array and calculates x, y, speed and heading for every
	 * waypoint.
	 */
	private void calculateWaypointDerivedValues ()
	{

		// determine min/max lat and long
		minLat = Double.MAX_VALUE;
		maxLat = -Double.MAX_VALUE;
		minLong = Double.MAX_VALUE;
		maxLong = -Double.MAX_VALUE;

		for (int i=0; i<waypoint.length; i++)
		{
			if (waypoint[i].coord.x > maxLong) maxLong = waypoint[i].coord.x;
			if (waypoint[i].coord.y > maxLat) maxLat = waypoint[i].coord.y;
			if (waypoint[i].coord.x < minLong) minLong = waypoint[i].coord.x;
			if (waypoint[i].coord.y < minLat) minLat = waypoint[i].coord.y;
		}

		// calculate x, y, speed, and heading for all members
		double latFactor = 111120;
		double medianLatitude = (maxLat + minLat) / 2;
		double longFactor = latFactor * Math.cos (Math.toRadians (medianLatitude));

		for (int i=0; i<waypoint.length; i++)
		{
			Waypoint to = waypoint[i];
			to.x = (int) ((to.coord.x - minLong) * longFactor);
			to.y = (int) ((maxLat - to.coord.y) * latFactor);
			
			if (i > 0)
			{
				Waypoint from = waypoint[i-1];
				int xd = to.x-from.x;
				int yd = to.y-from.y;
				double distance = Math.sqrt(xd*xd+yd*yd);
				if (to.time == from.time)
				{
					to.speed = 0;
				}
				else
				{
					to.speed = (int) (distance * 3600 / (double) (to.time - from.time));
				}

				if (to.y == from.y)
				{
					to.heading = 0;
				}
				else
				{
					double cos = -((double) yd) / distance;
					double phi = Math.acos (cos);

					if (xd < 0) phi = 2 * Math.PI - phi;

					to.heading = (int) Math.toDegrees (phi);
					if (to.heading < 0) to.heading += 360;
				}
			}
		}
		waypoint[0].heading = waypoint[1].heading;
		waypoint[0].speed = waypoint[1].speed;

		for (int i=0; i<waypoint.length-1; i++)
		{
			waypoint[i].speed = (waypoint[i].speed + waypoint[i+1].speed) / 2;

/*
			System.out.println ("\tlat/long: "+waypoint[i].coord.y+"/"+waypoint[i].coord.x);
			System.out.println ("\tx/y:      "+waypoint[i].x+"/"+waypoint[i].y);
			System.out.println ("Waypoint "+i);
			System.out.println ("\ttime:     "+waypoint[i].time);
			System.out.println ("\tspd/hdg:  "+waypoint[i].speed+"/"+waypoint[i].heading);
*/
		}
			
		edgeEdit.updateMap ();
	}

	/**
	 * Opens a file chooser for JPEG images and calls loadPhotos (Files[]) with the
	 * result.
	 */
	public void loadPhotos ()
	{
		JFileChooser chooser = new JFileChooser (new File (System.getProperty ("user.dir")));
		chooser.setMultiSelectionEnabled (true);
		chooser.setFileFilter (new javax.swing.filechooser.FileFilter () {
			public String getDescription () { return "JPEG images"; };
			public boolean accept (File f) { 
				if (f.isDirectory ()) return true;
				String name = f.getName ();
				int dot = name.lastIndexOf (".");
				if (dot < 1) return false;
				String ext = name.substring (dot + 1).toLowerCase ();
				return (ext.equals ("jpg") || ext.equals ("jpeg"));
			}
		});
		int returnVal = chooser.showOpenDialog (this);
    	if (returnVal != JFileChooser.APPROVE_OPTION) 
		{
			return;
		}

		loadPhotos (chooser.getSelectedFiles ());
		
	}
	
	/**
	 * Populates the photo list with the given list of images. Calls "jhead"
	 * on each to find out the real timestamp. Also generates all thumbnails.
	 */
	public void loadPhotos (final File[] list)
	{

		final Runtime currentRuntime = Runtime.getRuntime ();
		final ProgressMonitor progressMonitor = new ProgressMonitor(this, 
			"Analyzing photos", "", 0, list.length);
		progressMonitor.setMillisToDecideToPopup (1000);
		loadPhotos.setEnabled (false);
		photoLocation.setEnabled (false);

		/* execute loading of photos in a separate thread */
		Thread subThread = new Thread (new Runnable () {
			public void run () 
			{
				Photo[] tmpPhoto = new Photo[list.length];
				String cmdArray[] = new String[2];
				String line;
				Process jhead;
				SimpleDateFormat jheadDF = new SimpleDateFormat("yyyy:MM:dd HH:mm:ss");
				cmdArray[0] = "jhead";
				for (int i=0; i<list.length; i++)
				{
					tmpPhoto[i] = new Photo ();
					try
					{
						tmpPhoto[i].filename = list[i].getCanonicalPath ();
						cmdArray[1] = tmpPhoto[i].filename;
						progressMonitor.setProgress (i);
						progressMonitor.setNote ("jhead " + tmpPhoto[i].filename);
						jhead = currentRuntime.exec (cmdArray);
						BufferedReader stdout = new BufferedReader 
							(new InputStreamReader (jhead.getInputStream ()));
						while ((line = stdout.readLine ()) != null)
						{
							if (line.startsWith ("Date/Time "))
							{
								tmpPhoto[i].time = jheadDF.parse (line.substring (15), 
									new ParsePosition (0)).getTime ();
								break;
							}
						}
/*
Resolution   : 512 x 384
*/
						stdout.close ();
					}
					catch (IOException io)
					{
						errorMessage ("Error executiong jhead subprocess", io);
						loadPhotos.setEnabled (true);
						return;
					}
					try
					{
						jhead.waitFor ();
					}
					catch (InterruptedException e)
					{
						/* ignore for now */
					}
				}
				progressMonitor.close ();
	
			// fixme?	currentJourney.timeCorrection = waypoint[0].time - tmpPhoto[0].time;
				alignPhotosWithTrack (tmpPhoto);
				currentJourney.photo = tmpPhoto;
				edgeEdit.updateJourney ();
				photoTableModel.refresh ();
				if (mode == MODE_PHOTO) mapWithTrack.repaint ();
				loadPhotos.setEnabled (true);
				photoLocation.setEnabled (true);
				loadThumbnails ();
			}
		});
		subThread.start ();
	}

	public void loadThumbnails ()
	{
		Thread subThread = new Thread (new Runnable () {
			public void run () 
			{
				for (int i=0; i<currentJourney.photo.length; i++)
				{
					currentJourney.photo[i].generateThumbnail ();
					if (mode == MODE_PHOTO)
					{
						photoTable.repaint ();
					}
				}
			}
		});
		subThread.setPriority (Thread.MIN_PRIORITY);
		subThread.start ();
	}

	/**
	 * Deletes currently selected images from the photo list.
	 */
	private void editCut ()
	{
		
		int selRow = photoTable.getSelectedRow ();
		int selCount = photoTable.getSelectedRowCount ();

		if (selCount == currentJourney.photo.length)
		{
			currentJourney.photo = null;
			photoLocation.setEnabled (false);
		}
		else
		{
			Photo[] tmpPhoto = new Photo[currentJourney.photo.length - selCount];
			if (selRow != 0)
				System.arraycopy (currentJourney.photo, 0, tmpPhoto, 0, selRow);
			if (selRow + selCount < currentJourney.photo.length)
				System.arraycopy (currentJourney.photo, selRow + selCount, tmpPhoto, selRow, 
					currentJourney.photo.length - selRow - selCount);
			currentJourney.photo = tmpPhoto;
		}
		photoTableModel.refresh ();
		if (mode == MODE_PHOTO) mapWithTrack.repaint ();
		edgeEdit.updateJourney ();
	}

	/**
	 * Allow user to enter a new photo time correction (difference between
	 * photo timestamp and GPS timestamp)
	 */
	public void photoLocationMapping ()
	{

		PhotoLocationAdjustmentDialog pd = new PhotoLocationAdjustmentDialog (this);
		pd.setVisible (true);
		if (!pd.isCancelled ())
		{
			if (pd.getValue () != currentJourney.timeCorrection)
			{
				currentJourney.timeCorrection = pd.getValue ();
				alignPhotosWithTrack ();
			}	
		}
		edgeEdit.updateJourney ();
	}

	/* operates on the global "photo" array */
	private void alignPhotosWithTrack ()
	{
		alignPhotosWithTrack (currentJourney.photo);
		photoTableModel.refresh ();
		if (mode == MODE_PHOTO) mapWithTrack.repaint ();
	}

	/* operates on any array */
	private void alignPhotosWithTrack (Photo[] photoArray)
	{
		if (photoArray == null) return;
		if (waypoint == null) return;

		for (int i=0; i<photoArray.length; i++)
		{
			Waypoint result = new Waypoint ();
			result.time = photoArray[i].time + currentJourney.timeCorrection;
			int j;
			for (j=1; j<waypoint.length-1; j++)
			{
				if (waypoint[j].time >= result.time) break;
			}
			Waypoint from = waypoint[j-1];
			Waypoint to = waypoint[j];
			/* 0 <= portion <= 1 */
			double portion = ((double) (result.time - from.time)) / 
			                 ((double) (to.time - from.time));
			result.coord = new Point2D.Double 
				(from.coord.x + (to.coord.x - from.coord.x) * portion,
				from.coord.y + (to.coord.y - from.coord.y) * portion);

			result.x = from.x + (int) (((double) (to.x - from.x)) * portion);
			result.y = from.y + (int) (((double) (to.y - from.y)) * portion);
			result.speed = from.speed + (int) (((double) (to.speed - from.speed)) * portion);
			result.heading = to.heading;
			if (currentEdge.map != null) result.determineMapPixel (currentEdge.map);
			
			photoArray[i].waypoint = result;
		}
	}

	public static void main(String args[])
	{
		JourneyMaker app = new JourneyMaker ();
		app.setVisible (true);

		Vector xmlFiles = new Vector ();
		
		for (int i=0; i<args.length; i++)
		{
			if (args[i].startsWith ("-v"))
			{
				String vfn = null;
				if ((args[i].length () == 2) && (i < args.length - 1))
				{
					vfn = args[++i];
				}
				else if (args[i].length () > 2)
				{
					vfn = args[i].substring (2);
				}
				if (vfn != null)
				{
					app.loadVertices (new File (vfn));
				}
			}
			else if (args[i].startsWith ("-t"))
			{
				String tfn = null;
				if ((args[i].length () == 2) && (i < args.length - 1))
				{
					tfn = args[++i];
				}
				else if (args[i].length () > 2)
				{
					tfn = args[i].substring (2);
				}
				if (tfn != null)
				{
					app.loadTrack (new File (tfn));
				}
			}
			else
			{
				xmlFiles.addElement (new File (args[i]));
			}
		}
					
		if (!xmlFiles.isEmpty ())
		{
			File files[] = new File[xmlFiles.size ()];
			xmlFiles.copyInto (files);
			app.loadXML (files);
		}
/*
		x.loadTrack(new File("/tmp/road/track"));
		x.loadMap(new File("/tmp/road/map.gif"));
*/
/*
		x.loadPhotos(new File("/tmp/road/").listFiles (new java.io.FilenameFilter() {
			public boolean accept(File dir, String name)
			{
				int dot = name.lastIndexOf (".");
				if (dot < 1) return false;
				String ext = name.substring (dot + 1).toLowerCase ();
				return (ext.equals ("jpg") || ext.equals ("jpeg"));
			}
		}));
*/
	}

	private class MapWithTrack extends JPanel 
	{

		private Map map = null;
		private BasicStroke dashedStroke = new BasicStroke (1.0f, BasicStroke.CAP_SQUARE, 
			BasicStroke.JOIN_MITER, 10.0f, new float[] { 3.0f, 3.0f }, 0.0f);
		private BasicStroke wideStroke = new BasicStroke (4.0f);
		private BasicStroke plainStroke = new BasicStroke ();
		private ImageIcon mapImageIcon = null;
		private double scale = 1;
		private int mouseSelectIndex = -1;
		private Point mouseSelectPosition = null;
		private Point currentHack = null;

		public MapWithTrack ()
		{
			addMouseListener (new MouseListener () {
				public void mousePressed (MouseEvent e) { mousePress (e.getPoint (),
					(e.getModifiers () & InputEvent.BUTTON3_MASK) > 0); }
				public void mouseReleased (MouseEvent e) { mouseRelease (e.getPoint ()); }
				public void mouseClicked (MouseEvent e) { /* ignore */ }
				public void mouseEntered (MouseEvent e) { /* ignore */ }
				public void mouseExited (MouseEvent e) { mouseExit (); }
			});
			addMouseMotionListener (new MouseMotionListener () {
				public void mouseMoved (MouseEvent e) { mouseMove (e.getPoint ()); }
				public void mouseDragged (MouseEvent e) { mouseMove (e.getPoint ()); }
			});
		}

		public boolean setMap (Map m)
		{
			map = m;
			String file = (m.draftfilename != null) ? m.draftfilename : m.filename;
			mapImageIcon = new ImageIcon (file);
			while (mapImageIcon.getImageLoadStatus () == MediaTracker.LOADING);
			if (mapImageIcon.getImageLoadStatus () != MediaTracker.COMPLETE)
			{
				return false;
			}
			invalidate ();
			repaint ();
//setScale (40);
			return true;
		}

		private int findClickedPointIndex (Point pos)
		{
			Rectangle tempRect = new Rectangle ((int) ((pos.x -3 )/ scale), 
				(int) ((pos.y -3) / scale), (int) Math.max (2, 6 / scale),
				(int) Math.max (2, 6 / scale));

			if (mode == MODE_TRACK)
			{
				if (waypoint == null) return -1;
				for (int i = 0; i < waypoint.length; i++)
				{
					if (tempRect.contains (waypoint[i].mapPixel)) return i;
				}
			}
			else if (mode == MODE_PHOTO)
			{
				if (currentJourney.photo == null) return -1;
				for (int i = 0; i < currentJourney.photo.length; i++)
				{
					if (tempRect.contains (currentJourney.photo[i].waypoint.mapPixel)) return i;
				}
			}
			return -1;
		}

		private void mousePress (Point pos, boolean rightClick)
		{

			if ((mode == MODE_PHOTO) && (currentJourney.photo == null)) return;
			mouseSelectIndex = findClickedPointIndex (pos);
			if (mouseSelectIndex > -1)
			{
				Waypoint w = currentJourney.photo[mouseSelectIndex].waypoint;
				repaint ((int) (w.mapPixel.x * scale) - 5, 
					(int) (w.mapPixel.y * scale) - 5, 10, 10);
				currentHack = w.hack;
			}
			else
			{
				currentHack = null;
			}
			mouseSelectPosition = pos;
		
		}

		public void mouseMove (Point pos) 
		{

			if ((mouseSelectIndex > -1) && (mode == MODE_PHOTO))
			{
				Waypoint w = currentJourney.photo[mouseSelectIndex].waypoint;
				repaint ((int) (w.mapPixel.x * scale) - 5, 
					(int) (w.mapPixel.y * scale) - 5, 10, 10);
				
				int xoff = (int) ((pos.x - mouseSelectPosition.x) / scale);
				if (currentHack != null) xoff += currentHack.x;
				int yoff = (int) ((pos.y - mouseSelectPosition.y) / scale);
				if (currentHack != null) yoff += currentHack.y;

				if (w.hack != null)
				{
					w.hack.setLocation (xoff, yoff);
				}
				else
				{
					w.hack = new Point (xoff, yoff);
				}
				w.determineMapPixel (currentEdge.map);

				repaint ((int) (w.mapPixel.x * scale) - 5, 
					(int) (w.mapPixel.y * scale) - 5, 10, 10);
			}

			Point2D.Double tmp = new Point2D.Double ();
			Waypoint wp = new Waypoint ();
			wp.coord = new Point2D.Double ();

			// position label
			try
			{
				map.getTransform ().inverseTransform (pos, tmp);
				wp.coord.y = tmp.y; wp.coord.x = tmp.x;
				coordLabel.setText (wp.prettyLatitude () + " " + wp.prettyLongitude ());
				coordLabel.repaint ();
			}
			catch (Exception ex)
			{
				/* ignore */
			}

		}

		public void mouseExit () 
        {
			coordLabel.setText (" ");
			coordLabel.invalidate ();
		}
	
		public void mouseRelease (Point pos) 
		{
			// mouseMove (pos);
			if (mouseSelectIndex > -1)
			{
				((mode == MODE_TRACK) ? waypointTable : photoTable).setRowSelectionInterval (mouseSelectIndex, mouseSelectIndex);
				if (mode == MODE_PHOTO)
				{
					Waypoint w = currentJourney.photo[mouseSelectIndex].waypoint;
					repaint ((int) (w.mapPixel.x * scale) - 5, 
						(int) (w.mapPixel.y * scale) - 5, 10, 10);
				}
			}
			mouseSelectIndex = -1;
		}

		public void setScale (int percent)
		{
			scale = ((double) percent) / 100.0;
			invalidate ();
			repaint ();
		}

		public void paintComponent (Graphics g)
		{
			int selectedX=-1, selectedY=-1;
			int msX = -1, msY = -1;
	
			super.paintComponent (g);
			if (mapImageIcon == null) return;
			Color dotColor = (mode == MODE_TRACK) ? Color.green : Color.yellow;

			g.drawImage (mapImageIcon.getImage (), 0, 0, 
				(int) (map.size.width * scale + 0.5), 
				(int) (map.size.height * scale + 0.5), 
				mapImageIcon.getImageObserver ());

			if (mode == MODE_PHOTO)
			{
			
				Photo[] array = currentJourney.photo;
				if (array == null) return;
				int lastX = 0, lastY = 0;
				int selectedIndex = photoTable.getSelectedRow ();

				// first draw un-selected items and lines
				for (int i=0; i<array.length; i++)
				{
					if (array[i].waypoint.mapPixel == null) return;
					int x = (int) (array[i].waypoint.mapPixel.x * scale + 0.5);
					int y = (int) (array[i].waypoint.mapPixel.y * scale + 0.5);
					if (!array[i].selected)
					{
						g.setColor (Color.gray);
						g.fillOval (x - 2, y - 2, 4, 4);
					}
					if ((i>0) && (!LINES_ONLY_IF_SELECTED || 
						(array[i].selected && array[i-1].selected)))
					{
						g.setColor (Color.black);
						g.drawLine (x, y, lastX, lastY);
					}
					lastX = x; 
					lastY = y;
					if (i == selectedIndex)
					{
						selectedX = x;
						selectedY = y;
					}
					if (i == mouseSelectIndex)
					{
						msX = x;
						msY = y;
					}
				}

				// then draw selected items
				for (int i=0; i<array.length; i++)
				{
					if (array[i].selected)
					{
						int x = (int) (array[i].waypoint.mapPixel.x * scale + 0.5);
						int y = (int) (array[i].waypoint.mapPixel.y * scale + 0.5);
						{
							g.setColor (Color.black);
							g.fillOval (x - 4, y - 4, 8, 8);
							g.setColor (Color.yellow);
							g.fillOval (x - 3, y - 3, 6, 6);
						}
					}
				}
			}
			else if (mode == MODE_TRACK)
			{
				if (waypoint == null) return;
				int selectedIndex = waypointTable.getSelectedRow ();
				int lastX = 0, lastY = 0;

				// first lines
				for (int i=0; i<waypoint.length; i++)
				{
					if (waypoint[i].mapPixel == null) return;
					int x = (int) (waypoint[i].mapPixel.x * scale + 0.5);
					int y = (int) (waypoint[i].mapPixel.y * scale + 0.5);
					if (i>0)
					{
						g.setColor (Color.black);
						g.drawLine (x, y, lastX, lastY);
					}
					lastX = x; 
					lastY = y;
					if (i == selectedIndex)
					{
						selectedX = x;
						selectedY = y;
					}
				}

				// then dots
				for (int i=0; i<waypoint.length; i++)
				{
					int x = (int) (waypoint[i].mapPixel.x * scale + 0.5);
					int y = (int) (waypoint[i].mapPixel.y * scale + 0.5);
					g.setColor (Color.black);
					g.fillOval (x - 4, y - 4, 8, 8);
					g.setColor (Color.green);
					g.fillOval (x - 3, y - 3, 6, 6);
				}
			}
			if (selectedX > -1)
			{
				g.setColor (Color.blue);
				g.fillOval (selectedX-3, selectedY-3, 6, 6);
			}
			if (msX > -1)
			{
				g.setColor (Color.red);
				g.fillOval (selectedX-3, selectedY-3, 6, 6);
			}

			if (mode == MODE_EDGE)
			{
				if (currentJourney.photo != null)
				{
					int firstSelected = -1, lastSelected = -1;
					for (int i=0; i < currentJourney.photo.length; i++)
					{
						if (currentJourney.photo[i].selected)
						{
							if (firstSelected == -1)	
							{
								firstSelected = i;
							}
							else
							{
								lastSelected = i;
							}
						}
					}
					if ((firstSelected > -1) && (lastSelected > -1))
					{
						Point from = new Point (currentJourney.photo[firstSelected].
							waypoint.mapPixel);
						Point to = new Point (currentJourney.photo[lastSelected].
							waypoint.mapPixel);
						scale (from);
						scale (to);
						((Graphics2D) g).setStroke (wideStroke);
						g.drawLine (from.x, from.y, to.x, to.y);
						((Graphics2D) g).setStroke (plainStroke);
						GraphUtil.fill_arrowhead ((Graphics2D) g, from.x, from.y, to.x, to.y,
							30, 30, 0.80);
						g.setColor (Color.black);
						g.fillOval (from.x-dotSize, from.y-dotSize, 2*dotSize, 2*dotSize);
						g.fillOval (to.x-dotSize, to.y-dotSize, 2*dotSize, 2*dotSize);
						g.setColor (Color.yellow);
						g.fillOval (from.x-dotSize+1, from.y-dotSize+1, 2*dotSize-2, 2*dotSize-2);
						g.fillOval (to.x-dotSize+1, to.y-dotSize+1, 2*dotSize-2, 2*dotSize-2);
					}
				}
			}


		}

		private void scale (Point p)
		{
			p.x = (int) (p.x * scale + 0.5);
			p.y = (int) (p.y * scale + 0.5);
		}
	
		public Dimension getPreferredSize ()
		{
			if (mapImageIcon != null)
			{
				return (new Dimension ((int)(map.size.width*scale + .5), 
					(int)(map.size.height*scale+.5)));
			}
			return (new Dimension (200, 200));
		}
	
		
	}

	private class WaypointTableModel extends ReadOnlyTableModel
	{

		private	SimpleDateFormat timeFmt = new SimpleDateFormat("HH:mm:ss");
		private FieldPosition zero = new FieldPosition (0);
		private Date dateObject = new Date ();
		private StringBuffer timeStringBuffer = new StringBuffer ();


		public Class getColumnClass (int index)
		{
			switch (index)
			{
				case 0: return Integer.class;
				case 1: 
				case 2: 
				case 3: return String.class;
				case 4:
				case 5: return Integer.class;
			}
			return null;
		}

		public int getColumnCount ()
		{
			return 6;
		}

		public String getColumnName (int index)
		{
			switch (index)
			{
				case 0: return "Index";
				case 1: return "Time";
				case 2: return "Latitude";
				case 3: return "Longitude";
				case 4: return "Speed";
				case 5: return "Heading";
			}
			return null;
		}
		
		public int getRowCount ()
		{
			return (waypoint == null) ? 0 : waypoint.length;
		}

		public Object getValueAt (int row, int col)
		{
			Waypoint wp = waypoint[row];
			switch (col)
			{
				case 0: return new Integer (row+1);
				case 1: dateObject.setTime (wp.time);
						timeStringBuffer.setLength (0);
						return (timeFmt.format (dateObject, timeStringBuffer, zero));
				case 2: return wp.prettyLatitude ();
				case 3: return wp.prettyLongitude ();
				case 4: return new Integer (wp.speed);
				case 5: return new Integer (wp.heading);
			}
			return null;
		}
		
	}

	private class VertexTableModel extends ReadOnlyTableModel
	{

		private StringBuffer formatHelper = new StringBuffer ();

		public Class getColumnClass (int index)
		{
			switch (index)
			{
				case 0: return Integer.class;
				case 1: 
				case 2: 
				case 3:
				case 4: return String.class;
			}
			return null;
		}

		public int getColumnCount ()
		{
			return 5;
		}

		public String getColumnName (int index)
		{
			switch (index)
			{
				case 0: return "Index";
				case 1: return "Id";
				case 2: return "Name";
				case 3: return "Connections";
				case 4: return "Detail";
			}
			return null;
		}
		
		public int getRowCount ()
		{
			return (vertex == null) ? 0 : vertex.length;
		}

		public Object getValueAt (int row, int col)
		{
			Vertex vx = vertex[row];
			switch (col)
			{
				case 0: return new Integer (row+1);
				case 1: return vx.id;
				case 2: return vx.name;
				case 3: formatHelper.setLength (0);
						if (vx.connector == null) return null;
						for (int i=0; i<vx.connector.length; i++)
						{
							formatHelper.append (vx.connector[i]);
							formatHelper.append (" (");
							formatHelper.append (vx.direction[i]);
							formatHelper.append (")");
							if (i<vx.connector.length-1) formatHelper.append(", ");
						}
						return formatHelper.toString ();
				case 4: return vx.getDetail ("en");
			}
			return null;
		}
		
	}

	private class PhotoTableModel extends ReadWriteTableModel
	{

		private	SimpleDateFormat timeFmt = new SimpleDateFormat("HH:mm:ss");
		private FieldPosition zero = new FieldPosition (0);
		private Date dateObject = new Date ();
		private StringBuffer timeStringBuffer = new StringBuffer ();


		public Class getColumnClass (int index)
		{
			switch (index)
			{
				case 0: return Boolean.class;
				case 1: return Integer.class;
				case 2: 
				case 3: return String.class;
				case 4:
				case 5:
				case 6: return Integer.class;
			}
			return null;
		}

		public int getColumnCount ()
		{
			return 7;
		}

		public String getColumnName (int index)
		{
			switch (index)
			{
				case 0: return "Selected";
				case 1: return "Index";
				case 2: return "Time";
				case 3: return "File";
				case 4: return "Brightness";
				case 5: return "Correction";
				case 6: return "Gamma";
			}
			return null;
		}

		public void setColumnWidths (JTable t)
		{
			for (int i=0; i<getColumnCount (); i++)
			{
				TableColumn c = t.getColumnModel ().getColumn (i);
				switch (i)
				{
				  case 0: c.setPreferredWidth (30); break;
				  case 1: c.setPreferredWidth (30); break;
				  case 2: c.setPreferredWidth (90); break;
				  case 3: c.setPreferredWidth (90); break;
				  case 4: c.setPreferredWidth (15); break;
				  case 5: c.setPreferredWidth (30); break;
				  case 6: c.setPreferredWidth (30); break;
				}
			}
			t.sizeColumnsToFit (-1);
		}

		public void setColumnRenderers (JTable t)
		{
			t.getColumnModel ().getColumn (4).setCellRenderer (new ColorRenderer ());
		}
			
		
		public int getRowCount ()
		{
			return (currentJourney.photo == null) ? 0 : currentJourney.photo.length;
		}

		public Object getValueAt (int row, int col)
		{
			Photo p = currentJourney.photo[row];
			Waypoint wp = p.waypoint;
			switch (col)
			{
				case 0: return new Boolean (p.selected);
				case 1: return new Integer (row+1);
				case 2: dateObject.setTime (wp.time);
						timeStringBuffer.setLength (0);
						return (timeFmt.format (dateObject, timeStringBuffer, zero));
				case 3: return p.filename.substring (p.filename.lastIndexOf("/") + 1);
				case 4: return new Integer (p.getAverageBrightness ());
				case 5: return new Integer (p.getBrightness ());
				case 6: return new Integer (p.getGamma ());
			}
			return null;
		}

	    public boolean isCellEditable (int row, int col)
		{
			return ((col == 0) || (col >= 5));
		}
		public void setValueAt (Object o, int row, int col)
		{
			if (col == 0)
			{
				currentJourney.photo[row].selected = ((Boolean)o).booleanValue ();
			}
			else if (col == 5)
			{
				currentJourney.photo[row].setBrightness (((Integer)o).intValue ());
				imagePanel.repaint ();
			}
			else if (col == 6)
			{
				currentJourney.photo[row].setGamma (((Integer)o).intValue ());
				imagePanel.repaint ();
			}
		}
		
	}

	private class ColorRenderer extends JLabel
	implements TableCellRenderer 
	{
		Border unselectedBorder = null;
		Border selectedBorder = null;

		public ColorRenderer () 
		{
			super ();
			setOpaque(true); 
        }

		public Component getTableCellRendererComponent(JTable table, Object number, 
			boolean isSelected, boolean hasFocus, int row, int column) 
		{
			int col = (int) ((double) (((Integer)number).intValue () - 128) * 1.25 + 128);
			if (col < 0) 
			{
				col = 0;
			}
			else if (col > 255)
			{
				col = 255;
			}

			setBackground (new Color (col, col, col ));
			if (isSelected) {
				if (selectedBorder == null) {
					selectedBorder = BorderFactory.createMatteBorder(2,5,2,5,
											  table.getSelectionBackground());
				}
				setBorder(selectedBorder);
			} else {
				if (unselectedBorder == null) {
					unselectedBorder = BorderFactory.createMatteBorder(2,5,2,5,
											  table.getBackground());
				}
				setBorder(unselectedBorder);
			}
            return this;
        }
    }

	private class ImagePanel extends JPanel 
	{
		private Photo photo = null;

		public ImagePanel ()
		{
			addMouseListener (new MouseAdapter () {
				public void mouseClicked (MouseEvent e)
				{
					if (photo != null)
					{
						new FullImage (photo).setVisible (true);
					}
				}
			});
		}

		public void paintComponent (Graphics g)
		{
			super.paintComponent (g);
			if (photo == null) return;

			double xf = ((double) Photo.THUMBWIDTH) / ((double) getSize ().width);
			double yf = ((double) Photo.THUMBHEIGHT) / ((double) getSize ().height);
			double factor = Math.max(xf, yf);
			int width = (int) (((double) Photo.THUMBWIDTH) / factor);
			int height = (int) (((double) Photo.THUMBHEIGHT) / factor);

			// g.drawImage (icon.getImage (), 0, 0, width, height, icon.getImageObserver ());
			g.drawImage (photo.getThumbnail (), 0, 0, width, height, null);

		}

		public void setPhoto (Photo p)
		{
			photo = p;
			repaint ();
		}

	}

	private class PhotoLocationAdjustmentDialog extends JDialog
	{
		boolean cancelled = false;
		JSlider slider; 

		public PhotoLocationAdjustmentDialog (Frame aFrame)
		{

			super (aFrame, true);

			final JOptionPane optionPane;
			int from = (int) ((waypoint[0].time - currentJourney.photo[0].time)/1000);
			int to = (int) ((waypoint[waypoint.length-1].time - currentJourney.photo[currentJourney.photo.length-1].time)/1000);
			int current = (int) (currentJourney.timeCorrection / 1000);

			if (to < from) 
			{
				int x = from; 
				from = to;
				to = x;
			}
			if (current < from) current = from;
			if (current > to) current = to;
	
			slider = new JSlider (from, to, current);
			slider.setPaintLabels (true);
			int labelInterval = (to - from) / 10;
			int startLabels = to;
			
			slider.setLabelTable (slider.createStandardLabels (to - from));
		
			final String enterString = "Enter";
			Object[] options = new Object[] { enterString, "Cancel" };

			FieldPosition zero = new FieldPosition (0);
			SimpleDateFormat timeFmt = new SimpleDateFormat("HH:mm:ss");
			StringBuffer ptr = new StringBuffer ("Photo time range: ");
		 	timeFmt.format (new Date (currentJourney.photo[0].time), ptr, zero);
			ptr.append (" to ");
		 	timeFmt.format (new Date (currentJourney.photo[currentJourney.photo.length-1].time), ptr, zero);
	
			StringBuffer wtr = new StringBuffer ("Waypoint time range: ");
		 	timeFmt.format (new Date (waypoint[0].time), wtr, zero);
			wtr.append (" to ");
		 	timeFmt.format (new Date (waypoint[waypoint.length-1].time), wtr, zero);
	
			final JTextField numEdit = new JTextField (Integer.toString (current), 10);
			slider.addChangeListener (new ChangeListener () {
				public void stateChanged (ChangeEvent e) {
					numEdit.setText (Integer.toString (slider.getValue ()));
				}
			});

			numEdit.addFocusListener (new FocusListener () {
				public void focusGained (FocusEvent e) { };
				public void focusLost (FocusEvent e) { 
					try {
						slider.setValue (Integer.parseInt (numEdit.getText()));
					} 
					catch (Exception ex) { };
				}
			});
		
			optionPane = new JOptionPane(new Object[] { ptr, wtr, 
				"New time correction value (in seconds):", slider, numEdit }, 
				JOptionPane.PLAIN_MESSAGE, 
				JOptionPane.YES_NO_OPTION, null, options, options[0]);
			setContentPane (optionPane);
			setTitle ("Time Correction");

			optionPane.addPropertyChangeListener (new PropertyChangeListener () {
				public void propertyChange (PropertyChangeEvent e) {
					cancelled = !optionPane.getValue().equals (enterString);
					setVisible(false);
		        }
			});
			pack ();
			setLocationRelativeTo (aFrame);

		}

		public boolean isCancelled () { return cancelled; }
		public int getValue () { return slider.getValue () * 1000; }
	}
	
	private class XMLLoader extends JDialog
	{

		JList mapList = null;
		JList edgeList = null;
		JList journeyList = null;
		HashMap vertices = new HashMap ();
		HashMap maps = new HashMap ();
		HashMap edges = new HashMap ();
		HashMap journeys = new HashMap ();
		private boolean error = false;
		
		public XMLLoader (Frame owner)
		{
			this (owner, null);
		}

		public XMLLoader (Frame owner, File[] files)
		{
			super (owner, true);
			setTitle ("Load XML data");
			
			if (files == null)
			{
				JFileChooser chooser = new JFileChooser (new File (System.getProperty ("user.dir")));
				chooser.setMultiSelectionEnabled (true);
				chooser.setFileFilter (new javax.swing.filechooser.FileFilter () {
					public String getDescription () { return "XML files"; };
					public boolean accept (File f) { 
						if (f.isDirectory ()) return true;
						String name = f.getName ();
						int dot = name.lastIndexOf (".");
						if (dot < 1) return false;
						String ext = name.substring (dot + 1).toLowerCase ();
						return (ext.equals ("xml") || ext.equals ("XML"));
					};
				});

				if (chooser.showOpenDialog (owner) != JFileChooser.APPROVE_OPTION) 
				{
					error = true;
					return;
				}
				chooser.setVisible (false);
				files = chooser.getSelectedFiles ();
			}
	
			if (files.length == 0)
			{
				error = true;
				return;
			}
	
			// readFile will need access to the vertices we already have
			Vertex oldArray[] = JourneyMaker.this.vertex;
			if (oldArray != null)
			{
				for (int i=0; i<oldArray.length; i++)
					vertices.put (oldArray[i].id, oldArray[i]);
			}

			// try to load & parse all selected files.
			for (int i = 0; i < files.length; i++)
			{
				try
				{
					XMLUtils.readFile (parser, files[i].getCanonicalPath (), vertices, maps, 
						edges, journeys, null, debug ? System.out : null, true, true);
				}
				catch (Exception ex)
				{
					errorMessage ("Error while loading '" + files[i].getName () + "'", ex);
					error = true;
					return;		
				}
			}
	
			getContentPane ().setLayout (new SimpleGridBag ());

			JPanel chooserPanel = new JPanel ();
			chooserPanel.setLayout (new SimpleGridBag ());
			chooserPanel.add (new JLabel ("Select which items to load..."), "w=3 a=w f=h wx=1");
	
			String[] mapIds = new String[maps.size () + 1];
			mapIds[0] = "<retain current>";
			int c=0;
			for (Iterator i = maps.keySet ().iterator (); i.hasNext (); ) 
				mapIds[++c] = (String) i.next();
			if (c != 0) Arrays.sort (mapIds, 1, c+1);
			JScrollPane sp;
			chooserPanel.add (sp = new JScrollPane (mapList = new JList (mapIds)), "nl i=5 wx=.3 wy=1 f=b");
			mapList.setSelectedIndex (0);
			sp.setBorder (BorderFactory.createTitledBorder ("Maps"));
	
			String[] edgeIds = new String[edges.size () + 1];
			edgeIds[0] = "<retain current>";
			c=0;
			for (Iterator i = edges.keySet ().iterator (); i.hasNext (); ) 
				edgeIds[++c] = (String) i.next();
			if (c != 0) Arrays.sort (edgeIds, 1, c+1);
			chooserPanel.add (sp = new JScrollPane (edgeList = new JList (edgeIds)), "i=5 wx=.3 wy=1 f=b");
			edgeList.setSelectedIndex (0);
			sp.setBorder (BorderFactory.createTitledBorder ("Edges"));
			edgeList.addListSelectionListener (new ListSelectionListener () {
				public void valueChanged (ListSelectionEvent lse) { 
					if (edgeList.getSelectedIndex () > 0) {
						Edge e = (Edge) edges.get (edgeList.getSelectedValue ());
						mapList.setSelectedValue (e.map.id, true); 
					}
				  }
			});
	
			String[] journeyIds = new String[journeys.size () + 1];
			journeyIds[0] = "<retain current>";
			c=0;
			for (Iterator i = journeys.keySet ().iterator (); i.hasNext (); ) 
				journeyIds[++c] = (String) i.next();
			if (c != 0) Arrays.sort (journeyIds, 1, c+1);
			chooserPanel.add (sp = new JScrollPane (journeyList = new JList (journeyIds)), "i=5 wx=.3 wy=1 f=b");
			journeyList.setSelectedIndex (0);
			sp.setBorder (BorderFactory.createTitledBorder ("Journeys"));
			journeyList.addListSelectionListener (new ListSelectionListener () {
				public void valueChanged (ListSelectionEvent lse) { 
					if (journeyList.getSelectedIndex () > 0) {
						Edge e = ((Journey)journeys.get(journeyList.getSelectedValue())).findEdge (edges.values ().iterator ());
						if (e != null) edgeList.setSelectedValue (e.getHandle (), true);
					}
				}
			});
	
			getContentPane ().add (chooserPanel, "f=b i=15");
	
			JPanel buttonPanel = new JPanel ();
			JButton okButton = new JButton ("Load");
			JButton cancelButton = new JButton ("Cancel");
			buttonPanel.add (cancelButton);
			buttonPanel.add (okButton);
			getContentPane ().add (buttonPanel, "nl a=e f=h wx=1");

			cancelButton.addActionListener (new ActionListener () {
				public void actionPerformed (ActionEvent e)
					{ mapList = null; edgeList = null; journeyList = null; setVisible (false); }
			});
			okButton.addActionListener (new ActionListener () {
				public void actionPerformed (ActionEvent e)
					{ setVisible (false); }
			});

			pack ();
			setLocationRelativeTo (owner);
		}

		public void addMap (Map m)
		{
			maps.put (m.id, m);
		}

		public Map getSelectedMap ()
		{
			if ((mapList == null) || (mapList.getSelectedIndex () == 0))
				return null;
			return  (Map) maps.get (mapList.getSelectedValue ());
		}
		public Edge getSelectedEdge ()
		{
			if ((edgeList == null) || (edgeList.getSelectedIndex () == 0))
				return null;
			return  (Edge) edges.get (edgeList.getSelectedValue ());
		}
		public Journey getSelectedJourney ()
		{
			if ((journeyList == null) || (journeyList.getSelectedIndex () == 0))
				return null;
			return  (Journey) journeys.get (journeyList.getSelectedValue ());
		}

		public boolean hasError ()
		{
			return error;
		}
	}

	private class FullImage extends JFrame
	{
		Image theImage = null;
		public FullImage (Photo p)
		{
			super (p.filename);
			setSize (1024, 768);
			theImage = Toolkit.getDefaultToolkit ().getImage (p.filename);
		}
		public void paint (Graphics g)
		{
			super.paint (g);
			g.drawImage (theImage, 0, 0, new java.awt.image.ImageObserver () {
				public boolean imageUpdate(Image img, int infoflags, int x, int y, int w, int h) 
				{ 
					if ((infoflags & java.awt.image.ImageObserver.ALLBITS) != 0)
					{
						FullImage.this.repaint ();
						return false;
					}
					/*
					if ((infoflags & java.awt.image.ImageObserver.SOMEBITS) != 0)
					{
						FullImage.this.repaint ();
					}
					*/
					return true;
				}
			});
		}
	}
}

