
package org.remote.roadhog;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;

import org.w3c.dom.*;

import magick.*;

public class Photo extends XMLObject
{
	public String filename;
	public long time;
	public int xres;
	public int yres;
	private int brightness = 0;
	private int gamma = 0;
	private volatile BufferedImage thumbnailImage;
	private volatile MagickImage thumbnailMagick;
	private int avbrightness = 0;

	public static int THUMBWIDTH=256;
	public static int THUMBHEIGHT=192;

	public volatile boolean selected = false;


	/* this is an interpolated, theoretical waypoint that lies between two 
	 * measured points
	 */
	public Waypoint waypoint;

	/**
	 * Creates empty Photo 
	 */
	public Photo ()
	{
		/* empty */
	}
	
	/**
	 * Creates new Photo based on XML node
	 */
	public Photo (Node node)
	throws ParseException
	{
		NodeList sublist;
		Element photo = (Element) node;

		sublist = photo.getElementsByTagName ("filename");
		filename = getTextChild (sublist.item(0));

		try
		{
			sublist = photo.getElementsByTagName ("time");
			time = Long.parseLong (getTextChild (sublist.item(0)));

			sublist = photo.getElementsByTagName ("width");
			xres = Integer.parseInt (getTextChild (sublist.item(0)));

			sublist = photo.getElementsByTagName ("height");
			yres = Integer.parseInt (getTextChild (sublist.item(0)));

			sublist = photo.getElementsByTagName ("brightness");
			if (sublist.getLength () > 0) 
				brightness = Integer.parseInt (getTextChild (sublist.item(0)));

			sublist = photo.getElementsByTagName ("gamma");
			if (sublist.getLength () > 0) 
				gamma = Integer.parseInt (getTextChild (sublist.item(0)));

		}
		catch (NumberFormatException nfe)
		{
			throw new ParseException ("Number format error");
		}

		sublist = photo.getElementsByTagName ("waypoint");
		waypoint = new Waypoint (sublist.item(0));
		
	}

	/**
	 * writes XML serialized version of photo to stream
	 */
	public void toXML (PrintStream o, String indent)
	{
		o.println (indent + "<photo>");
		o.println (indent + "   <filename>" + stringToXML (filename) + "</filename>");
		o.println (indent + "   <time>" + time + "</time>");
		o.println (indent + "   <width>" + xres + "</width>");
		o.println (indent + "   <height>" + yres + "</height>");
		o.println (indent + "   <brightness>" + brightness + "</brightness>");
		o.println (indent + "   <gamma>" + gamma + "</gamma>");
		if (waypoint != null)
			waypoint.toXML (o, indent + "   ");
		o.println (indent + "</photo>");
	}

	public Image getThumbnail ()
	{
		if (thumbnailImage == null)
		{
			generateThumbnail ();
		}
		return thumbnailImage;
	}

	public void generateThumbnail ()
	{
		
		try
		{
			ImageInfo ii = new ImageInfo (filename);
			MagickImage tmp = new MagickImage (ii);
			if (thumbnailMagick != null) thumbnailMagick.destroyImage ();

			if (gamma != 0)
			{
				tmp.gammaImage (Double.toString ((100.0 + gamma)/100.0));
			}
			if (brightness != 0)
			{
				tmp.modulateImage (Integer.toString (100 + brightness));
			}
			thumbnailMagick = tmp.scaleImage (THUMBWIDTH, THUMBHEIGHT);
			tmp.destroyImage ();
			if (thumbnailImage != null) thumbnailImage.flush ();
			thumbnailImage = new BufferedImage (THUMBWIDTH, THUMBHEIGHT, BufferedImage.TYPE_INT_RGB);
			Graphics2D gr = thumbnailImage.createGraphics ();
			gr.drawImage (Toolkit.getDefaultToolkit ().createImage (new MagickProducer (thumbnailMagick)), 0, 0, THUMBWIDTH, THUMBHEIGHT, null);

			avbrightness = calculateAverageBrightness (thumbnailImage);

		}
		catch (MagickException x)
		{
			x.printStackTrace ();
		}

	}

	public void setBrightness (int b)
	{
		if (b != brightness)
		{
			brightness = b;
			generateThumbnail ();
		}
	}

	public void setGamma (int g)
	{
		if (g != gamma)
		{
			gamma = g;
			generateThumbnail ();
		}
	}

	public int getBrightness ()
	{
		return brightness;
	}

	public int getGamma ()
	{
		return gamma;
	}

	private int calculateAverageBrightness (BufferedImage m)
	{
		int[] rgbarray = new int[THUMBWIDTH * THUMBHEIGHT];
		m.getRGB (0, 0, THUMBWIDTH, THUMBHEIGHT, rgbarray, 0, THUMBWIDTH);
		int r=0;
		int g=0;
		int b=0;
		for (int i=0; i<rgbarray.length; i++)
		{
			r += rgbarray[i]>>16 & 0xFF;
			g += rgbarray[i]>>8 & 0xFF;
			b += rgbarray[i] & 0xFF;
		}
		return (int) (0.299 * (double) r + 0.587 * (double) g + 0.114 * (double) b) / rgbarray.length;
	}

	public int getAverageBrightness ()
	{
		return avbrightness;
	}

	public JPanel makePanel ()
	{
		return (new PhotoPanel ());
	}

	private class PhotoPanel extends JPanel
	{
		public PhotoPanel ()
		{
			setLayout (null);
			setSize (THUMBWIDTH, THUMBHEIGHT);
			generateThumbnail ();
		}
		public void paint (Graphics g)
		{
			super.paint (g);
			((Graphics2D)g).drawImage (thumbnailImage, new AffineTransform (), null);
		}
	}

	private int byteToInt (byte b)
	{
		return (b>=0) ? b : b + 256;
	}

}
